#!/usr/bin/python
#
# Copyright (c) 2021 Aparna Patil(@aparna-patil)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_firewallpolicy

version_added: "1.13.0"

short_description: Create, delete or update specified firewall policy.

description:
    - Creates, deletes, or updates given firewall policy in specified resource group.

options:
    resource_group:
        description:
            - Name of the resource group.
        required: true
        type: str
    name:
        description:
            - The name of the firewall policy.
        required: true
        type: str
    location:
        description:
            - Location for firewall policy. Defaults to location of resource group if not specified.
        type: str
    base_policy:
        description:
            - The name of the parent firewall policy from which rules are inherited.
        type: str
    threat_intel_mode:
        description:
            - The operation mode for Threat Intel.
        default: alert
        type: str
        choices:
            - alert
            - deny
            - 'off'
    threat_intel_whitelist:
        description:
            - ThreatIntel Whitelist for Firewall Policy.
        type: dict
        suboptions:
            ip_addresses:
                description:
                    - List of IP addresses for the ThreatIntel Whitelist.
                type: list
                elements: str
            append_ip_addresses:
                description:
                    - Flag to indicate if the ip_addresses to be appended or not.
                type: bool
                default: true
            fqdns:
                description:
                    - List of FQDNs for the ThreatIntel Whitelist
                type: list
                elements: str
            append_fqdns:
                description:
                    - Flag to indicate if the fqdns to be appended or not.
                type: bool
                default: true
    state:
        description:
            - Assert the state of the firewall policy. Use C(present) to create or update and C(absent) to delete.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Aparna Patil (@aparna-patil)
'''

EXAMPLES = '''
- name: Create a Firewall Policy
  azure_rm_firewallpolicy:
    resource_group: myAzureResourceGroup
    name: myfirewallpolicy
    base_policy: firewallparentpolicy
    threat_intel_mode: alert
    threat_intel_whitelist:
      ip_addresses:
        - 10.0.0.1
        - 10.0.0.2
      fqdns:
        - "*.microsoft.com"
        - "*.azure.com"
    state: present

- name: Update Firewall Policy
  azure_rm_firewallpolicy:
    resource_group: myAzureResourceGroup
    name: myfirewallpolicy
    base_policy: firewallparentpolicy
    threat_intel_mode: deny
    threat_intel_whitelist:
      ip_addresses:
        - 10.0.0.1
      fqdns:
        - "*.microsoft.com"
    state: present
    tags:
      key1: "value1"

- name: Delete Firewall Policy
  azure_rm_firewallpolicy:
    resource_group: myAzureResourceGroup
    name: myfirewallpolicy
    state: absent
'''

RETURN = '''
state:
    description:
        - Current state of the Firewall Policy.
    returned: always
    type: complex
    contains:
        id:
            description:
                - The firewall policy ID.
            returned: always
            type: str
            sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyAzureResourceGroup/providers/
                     Microsoft.Network/firewallPolicies/myfirewallpolicy"
        name:
            description:
                - The firewall policy name.
            returned: always
            type: str
            sample: 'myfirewallpolicy'
        location:
            description:
                - The Azure Region where the resource lives.
            returned: always
            type: str
            sample: eastus
        base_policy:
            description:
                - The parent firewall policy from which rules are inherited.
            returned: always
            type: dict
            sample: {
              "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyAzureResourceGroup/providers/
                     Microsoft.Network/firewallPolicies/firewallparentpolicy"
            }
        child_policies:
            description:
                - List of references to Child Firewall Policies.
            returned: always
            type: list
            elements: dict
            sample: [
            {
              "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyAzureResourceGroup/providers/
                     Microsoft.Network/firewallPolicies/childpolicy1"
            }
        ]
        provisioning_state:
            description:
                - The provisioning state of the resource.
            returned: always
            type: str
            sample: Succeeded
        firewalls:
            description:
                - List of references to Azure Firewalls that this Firewall Policy is associated with.
            returned: always
            type: list
            elements: dict
            sample: [
            {
              "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myAzureResourceGroup/providers/
                     Microsoft.Network/azureFirewalls/azurefirewall"
            }
        ]
        rule_collection_groups:
            description:
                - List of references to FirewallPolicyRuleCollectionGroups.
            returned: always
            type: list
            elements: dict
            sample: [
            {
              "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyAzureResourceGroup/providers/
                     Microsoft.Network/firewallPolicies/myfirewallpolicy/
                     ruleCollectionGroups/DefaultNetworkRuleCollectionGroup"
            }
        ]
        threat_intel_mode:
            description:
                - The operation mode for Threat Intelligence.
            returned: always
            type: str
            sample: Alert
        threat_intel_whitelist:
            description:
                - ThreatIntel Whitelist for Firewall Policy.
            returned: always
            type: dict
            sample: {
              "fqdns": [
                  "*.microsoft.com",
                  "*.azure.com"
              ],
              "ip_addresses": [
                  "10.0.0.1",
                  "10.0.0.2"
              ]
        }
        tags:
            description:
                - Resource tags.
            returned: always
            type: list
            sample: [{"key1": "value1"}]
        type:
            description:
                - The type of resource.
            returned: always
            type: str
            sample: Microsoft.Network/FirewallPolicies
        etag:
            description:
                - The etag of the firewall policy.
            returned: always
            type: str
            sample: 7cb2538e-0e52-4435-8979-4f417e7269d1
'''

from ansible.module_utils.basic import _load_params
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase, \
    format_resource_id, normalize_location_name
import copy

try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.core.polling import LROPoller
except ImportError:
    # This is handled in azure_rm_common
    pass

threat_intel_whitelist_spec = dict(
    ip_addresses=dict(type='list', elements='str'),
    append_ip_addresses=dict(type='bool', default=True),
    fqdns=dict(type='list', elements='str'),
    append_fqdns=dict(type='bool', default=True)
)


class AzureRMFirewallPolicy(AzureRMModuleBase):

    def __init__(self):

        _load_params()
        # define user inputs from playbook
        self.module_arg_spec = dict(
            resource_group=dict(type='str', required=True),
            name=dict(type='str', required=True),
            location=dict(type='str'),
            base_policy=dict(type='str'),
            threat_intel_mode=dict(choices=['alert', 'deny', 'off'], default='alert', type='str'),
            threat_intel_whitelist=dict(type='dict', options=threat_intel_whitelist_spec),
            state=dict(choices=['present', 'absent'], default='present', type='str'),
        )

        self.results = dict(
            changed=False,
            state=dict()
        )

        self.resource_group = None
        self.name = None
        self.state = None
        self.location = None
        self.base_policy = None
        self.threat_intel_mode = None
        self.threat_intel_whitelist = None
        self.tags = None

        super(AzureRMFirewallPolicy, self).__init__(self.module_arg_spec,
                                                    supports_tags=True,
                                                    supports_check_mode=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        changed = False
        results = dict()
        firewall_policy_old = None
        firewall_policy_new = None
        update_ip_address = False
        update_fqdns = False

        # retrieve resource group to make sure it exists
        resource_group = self.get_resource_group(self.resource_group)
        if not self.location:
            # Set default location
            self.location = resource_group.location

        self.location = normalize_location_name(self.location)

        if self.base_policy:
            base_policy = self.parse_resource_to_dict(self.base_policy)
            self.base_policy = format_resource_id(val=base_policy['name'],
                                                  subscription_id=base_policy['subscription_id'],
                                                  namespace='Microsoft.Network',
                                                  types='firewallPolicies',
                                                  resource_group=base_policy['resource_group'])

        try:
            self.log('Fetching Firewall policy {0}'.format(self.name))
            firewall_policy_old = self.network_client.firewall_policies.get(self.resource_group, self.name)
            # serialize object into a dictionary
            results = self.firewallpolicy_to_dict(firewall_policy_old)
            if self.state == 'present':
                changed = False
                update_tags, results['tags'] = self.update_tags(results['tags'])
                if update_tags:
                    changed = True
                self.tags = results['tags']
                if self.base_policy is not None:
                    if ('base_policy' not in results and self.base_policy != "") or \
                            ('base_policy' in results and self.base_policy != results['base_policy']['id']):
                        changed = True
                        results['base_policy'] = self.base_policy
                if self.threat_intel_mode is not None and \
                        self.threat_intel_mode.lower() != results['threat_intel_mode'].lower():
                    changed = True
                    results['threat_intel_mode'] = self.threat_intel_mode
                if self.threat_intel_whitelist is not None:
                    if 'threat_intel_whitelist' not in results:
                        changed = True
                        results['threat_intel_whitelist'] = self.threat_intel_whitelist
                    else:
                        update_ip_addresses, results['threat_intel_whitelist']['ip_addresses'] = \
                            self.update_values(results['threat_intel_whitelist']['ip_addresses']
                                               if 'ip_addresses' in results['threat_intel_whitelist'] else [],
                                               self.threat_intel_whitelist['ip_addresses']
                                               if self.threat_intel_whitelist['ip_addresses'] is not None else [],
                                               self.threat_intel_whitelist['append_ip_addresses'])
                        update_fqdns, results['threat_intel_whitelist']['fqdns'] = \
                            self.update_values(results['threat_intel_whitelist']['fqdns']
                                               if 'fqdns' in results['threat_intel_whitelist'] else [],
                                               self.threat_intel_whitelist['fqdns']
                                               if self.threat_intel_whitelist['fqdns'] is not None else [],
                                               self.threat_intel_whitelist['append_fqdns'])
                        if update_ip_addresses:
                            changed = True
                        self.threat_intel_whitelist['ip_addresses'] = results['threat_intel_whitelist']['ip_addresses']
                        if update_fqdns:
                            changed = True
                        self.threat_intel_whitelist['fqdns'] = results['threat_intel_whitelist']['fqdns']
            elif self.state == 'absent':
                changed = True

        except ResourceNotFoundError:
            if self.state == 'present':
                changed = True
            else:
                changed = False

        self.results['changed'] = changed
        self.results['state'] = results

        if self.check_mode:
            return self.results

        if changed:
            if self.state == 'present':
                # create or update firewall policy
                firewall_policy_new = \
                    self.network_models.FirewallPolicy(location=self.location,
                                                       threat_intel_mode=self.threat_intel_mode)
                if self.base_policy:
                    firewall_policy_new.base_policy = \
                        self.network_models.FirewallPolicy(id=self.base_policy)
                if self.threat_intel_whitelist:
                    firewall_policy_new.threat_intel_whitelist = self.network_models.FirewallPolicyThreatIntelWhitelist(
                        ip_addresses=self.threat_intel_whitelist['ip_addresses'],
                        fqdns=self.threat_intel_whitelist['fqdns']
                    )
                if self.tags:
                    firewall_policy_new.tags = self.tags
                self.results['state'] = self.create_or_update_firewallpolicy(firewall_policy_new)

            elif self.state == 'absent':
                # delete firewall policy
                self.delete_firewallpolicy()
                self.results['state'] = 'Deleted'

        return self.results

    def create_or_update_firewallpolicy(self, firewall_policy):
        try:
            # create a firewall policy
            response = self.network_client.firewall_policies.begin_create_or_update(resource_group_name=self.resource_group,
                                                                                    firewall_policy_name=self.name,
                                                                                    parameters=firewall_policy)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.fail("Error creating or updating Firewall policy {0} - {1}".format(self.name, str(exc)))
        return self.firewallpolicy_to_dict(response)

    def delete_firewallpolicy(self):
        try:
            # delete a firewall policy
            response = self.network_client.firewall_policies.begin_delete(resource_group_name=self.resource_group,
                                                                          firewall_policy_name=self.name)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.fail("Error deleting Firewall policy {0} - {1}".format(self.name, str(exc)))
        return response

    def update_values(self, existing_values, param_values, append):
        # comparing input values with existing values for given parameter

        new_values = copy.copy(existing_values)
        changed = False

        # check add or update
        for item in param_values:
            if item not in new_values:
                changed = True
                new_values.append(item)
        # check remove
        if not append:
            for item in existing_values:
                if item not in param_values:
                    new_values.remove(item)
                    changed = True
        return changed, new_values

    def firewallpolicy_to_dict(self, firewallpolicy):
        result = firewallpolicy.as_dict()
        result['tags'] = firewallpolicy.tags
        return result


def main():
    AzureRMFirewallPolicy()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: route53_health_check
version_added: 5.0.0
short_description: Manage health-checks in Amazons Route53 DNS service
description:
  - Creates and deletes DNS Health checks in Amazons Route53 service.
  - Only the port, resource_path, string_match and request_interval are
    considered when updating existing health-checks.
  - This module was originally added to C(community.aws) in release 1.0.0.
options:
  state:
    description:
      - Specifies the action to take.
    choices: [ 'present', 'absent' ]
    type: str
    default: 'present'
  disabled:
    description:
      - Stops Route 53 from performing health checks.
      - See the AWS documentation for more details on the exact implications.
        U(https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/health-checks-creating-values.html)
      - Defaults to C(true) when creating a new health check.
    type: bool
    version_added: 2.1.0
    version_added_collection: community.aws
  ip_address:
    description:
      - IP address of the end-point to check. Either this or I(fqdn) has to be provided.
      - IP addresses must be publicly routable.
    type: str
  port:
    description:
      - The port on the endpoint on which you want Amazon Route 53 to perform
        health checks. Required for TCP checks.
    type: int
  type:
    description:
      - The type of health check that you want to create, which indicates how
        Amazon Route 53 determines whether an endpoint is healthy.
      - Once health_check is created, type can not be changed.
      - The CALCULATED choice was added in 6.3.0.
    choices: [ 'HTTP', 'HTTPS', 'HTTP_STR_MATCH', 'HTTPS_STR_MATCH', 'TCP', 'CALCULATED' ]
    type: str
  child_health_checks:
    description:
      - The child health checks used for a calculated health check.
      - This parameter takes in the child health checks ids.
    type: list
    elements: str
    version_added: 6.3.0
  health_threshold:
    description:
      - The minimum number of healthy child health checks for a calculated health check to be considered healthy.
    default: 1
    type: int
    version_added: 6.3.0
  resource_path:
    description:
      - The path that you want Amazon Route 53 to request when performing
        health checks. The path can be any value for which your endpoint will
        return an HTTP status code of 2xx or 3xx when the endpoint is healthy,
        for example the file /docs/route53-health-check.html.
      - Mutually exclusive with I(type='TCP').
      - The path must begin with a /
      - Maximum 255 characters.
    type: str
  fqdn:
    description:
      - Domain name of the endpoint to check. Either this or I(ip_address) has
        to be provided. When both are given the I(fqdn) is used in the C(Host:)
        header of the HTTP request.
    type: str
  string_match:
    description:
      - If the check type is HTTP_STR_MATCH or HTTP_STR_MATCH, the string
        that you want Amazon Route 53 to search for in the response body from
        the specified resource. If the string appears in the first 5120 bytes
        of the response body, Amazon Route 53 considers the resource healthy.
    type: str
  request_interval:
    description:
      - The number of seconds between the time that Amazon Route 53 gets a
        response from your endpoint and the time that it sends the next
        health-check request.
    default: 30
    choices: [ 10, 30 ]
    type: int
  failure_threshold:
    description:
      - The number of consecutive health checks that an endpoint must pass or
        fail for Amazon Route 53 to change the current status of the endpoint
        from unhealthy to healthy or vice versa.
      - Will default to C(3) if not specified on creation.
    choices: [ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ]
    type: int
  health_check_name:
    description:
      - Name of the Health Check.
      - Used together with I(use_unique_names) to set/make use of I(health_check_name) as a unique identifier.
    type: str
    required: False
    aliases: ['name']
    version_added: 4.1.0
    version_added_collection: community.aws
  use_unique_names:
    description:
      - Used together with I(health_check_name) to set/make use of I(health_check_name) as a unique identifier.
    type: bool
    required: False
    version_added: 4.1.0
    version_added_collection: community.aws
  health_check_id:
    description:
      - ID of the health check to be update or deleted.
      - If provided, a health check can be updated or deleted based on the ID as unique identifier.
    type: str
    required: False
    aliases: ['id']
    version_added: 4.1.0
    version_added_collection: community.aws
  measure_latency:
    description:
      - To enable/disable latency graphs to monitor the latency between health checkers in multiple Amazon Web Services regions and your endpoint.
      - Value of I(measure_latency) is immutable and can not be modified after creating a health check.
        See U(https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/monitoring-health-check-latency.html)
    type: bool
    required: False
    version_added: 5.4.0
author:
  - "zimbatm (@zimbatm)"
notes:
  - Support for I(tags) and I(purge_tags) was added in release 2.1.0.
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: Create a health-check for host1.example.com and use it in record
  amazon.aws.route53_health_check:
    state: present
    fqdn: host1.example.com
    type: HTTP_STR_MATCH
    resource_path: /
    string_match: "Hello"
    request_interval: 10
    failure_threshold: 2
  register: my_health_check

- amazon.aws.route53:
    action: create
    zone: "example.com"
    type: CNAME
    record: "www.example.com"
    value: host1.example.com
    ttl: 30
    # Routing policy
    identifier: "host1@www"
    weight: 100
    health_check: "{{ my_health_check.health_check.id }}"

- name: create a simple health check with health_check_name as unique identifier
  amazon.aws.route53_health_check:
    state: present
    health_check_name: ansible
    fqdn: ansible.com
    port: 443
    type: HTTPS
    use_unique_names: true

- name: create a TCP health check with latency graphs enabled
  amazon.aws.route53_health_check:
    state: present
    health_check_name: ansible
    fqdn: ansible.com
    port: 443
    type: HTTPS
    use_unique_names: true
    measure_latency: true

- name: Delete health-check
  amazon.aws.route53_health_check:
    state: absent
    fqdn: host1.example.com

- name: Update Health check by ID - update ip_address
  amazon.aws.route53_health_check:
    id: 12345678-abcd-abcd-abcd-0fxxxxxxxxxx
    ip_address: 1.2.3.4

- name: Update Health check by ID - update port
  amazon.aws.route53_health_check:
    id: 12345678-abcd-abcd-abcd-0fxxxxxxxxxx
    ip_address: 8080

- name: Delete Health check by ID
  amazon.aws.route53_health_check:
    state: absent
    id: 12345678-abcd-abcd-abcd-0fxxxxxxxxxx
"""

RETURN = r"""
health_check:
  description: Information about the health check.
  returned: success
  type: dict
  contains:
    action:
      description: The action performed by the module.
      type: str
      returned: When a change is or would be made.
      sample: 'updated'
    id:
      description: The Unique ID assigned by AWS to the health check.
      type: str
      returned: When the health check exists.
      sample: 50ec8a13-9623-4c66-9834-dd8c5aedc9ba
    health_check_version:
      description: The version number of the health check.
      type: int
      returned: When the health check exists.
      sample: 14
    health_check_config:
      description:
        - Detailed information about the health check.
        - May contain additional values from Route 53 health check
          features not yet supported by this module.
      type: dict
      returned: When the health check exists.
      contains:
        type:
          description: The type of the health check.
          type: str
          returned: When the health check exists.
          sample: 'HTTPS_STR_MATCH'
        failure_threshold:
          description:
            - The number of consecutive health checks that an endpoint must pass or fail for Amazon Route 53 to
              change the current status of the endpoint from unhealthy to healthy or vice versa.
          type: int
          returned: When the health check exists.
          sample: 3
        fully_qualified_domain_name:
          description: The FQDN configured for the health check to test.
          type: str
          returned: When the health check exists and an FQDN is configured.
          sample: 'updated'
        ip_address:
          description: The IPv4 or IPv6 IP address of the endpoint to be queried.
          type: str
          returned: When the health check exists and a specific IP address is configured.
          sample: ''
        port:
          description: The port on the endpoint that the health check will query.
          type: str
          returned: When the health check exists.
          sample: 'updated'
        request_interval:
          description: The number of seconds between health check queries.
          type: int
          returned: When the health check exists.
          sample: 30
        resource_path:
          description: The URI path to query when performing an HTTP/HTTPS based health check.
          type: str
          returned: When the health check exists and a resource path has been configured.
          sample: '/healthz'
        search_string:
          description: A string that must be present in the response for a health check to be considered successful.
          type: str
          returned: When the health check exists and a search string has been configured.
          sample: 'ALIVE'
        disabled:
          description: Whether the health check has been disabled or not.
          type: bool
          returned: When the health check exists.
          sample: false
    tags:
      description: A dictionary representing the tags on the health check.
      type: dict
      returned: When the health check exists.
      sample: '{"my_key": "my_value"}'
"""

import uuid

try:
    import botocore
except ImportError:
    pass  # Handled by HAS_BOTO

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.route53 import get_tags
from ansible_collections.amazon.aws.plugins.module_utils.route53 import manage_tags


def _list_health_checks(**params):
    try:
        results = client.list_health_checks(aws_retry=True, **params)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to list health checks")
    return results


def find_health_check(ip_addr, fqdn, hc_type, request_interval, port):
    """Searches for health checks that have the exact same set of immutable values"""

    # In lieu of an Id we perform matches against the following values:
    # - ip_addr
    # - fqdn
    # - type (immutable)
    # - request_interval
    # - port

    # Because the list and route53 provides no 'filter' mechanism,
    # the using a paginator would result in (on average) double the
    # number of API calls and can get really slow.
    # Additionally, we can't properly wrap the paginator, so retrying means
    # starting from scratch with a paginator
    results = _list_health_checks()
    while True:
        for check in results.get("HealthChecks"):
            config = check.get("HealthCheckConfig")
            if (
                config.get("IPAddress", None) == ip_addr
                and config.get("FullyQualifiedDomainName", None) == fqdn
                and config.get("Type") == hc_type
                and config.get("RequestInterval") == request_interval
                and config.get("Port", None) == port
            ):
                return check

        if results.get("IsTruncated", False):
            results = _list_health_checks(Marker=results.get("NextMarker"))
        else:
            return None


def get_existing_checks_with_name():
    results = _list_health_checks()
    health_checks_with_name = {}
    while True:
        for check in results.get("HealthChecks"):
            if "Name" in describe_health_check(check["Id"])["tags"]:
                check_name = describe_health_check(check["Id"])["tags"]["Name"]
                health_checks_with_name[check_name] = check
        if results.get("IsTruncated", False):
            results = _list_health_checks(Marker=results.get("NextMarker"))
        else:
            return health_checks_with_name


def delete_health_check(check_id):
    if not check_id:
        return False, None

    if module.check_mode:
        return True, "delete"

    try:
        client.delete_health_check(
            aws_retry=True,
            HealthCheckId=check_id,
        )
    except is_boto3_error_code("NoSuchHealthCheck"):
        # Handle the deletion race condition as cleanly as possible
        return False, None
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Failed to list health checks")

    return True, "delete"


def create_health_check(
    ip_addr_in, fqdn_in, type_in, request_interval_in, port_in, child_health_checks_in, health_threshold_in
):
    # In general, if a request is repeated with the same CallerRef it won't
    # result in a duplicate check appearing.  This means we can safely use our
    # retry decorators
    caller_ref = str(uuid.uuid4())
    missing_args = []

    health_check = dict(
        Type=type_in,
    )
    if module.params.get("disabled") is not None:
        health_check["Disabled"] = module.params.get("disabled")
    if ip_addr_in:
        health_check["IPAddress"] = ip_addr_in
    if fqdn_in:
        health_check["FullyQualifiedDomainName"] = fqdn_in
    if port_in:
        health_check["Port"] = port_in

    if type_in in ["HTTP", "HTTPS", "HTTP_STR_MATCH", "HTTPS_STR_MATCH"]:
        resource_path = module.params.get("resource_path")
        # if not resource_path:
        #     missing_args.append('resource_path')
        if resource_path:
            health_check["ResourcePath"] = resource_path
    if type_in in ["HTTP_STR_MATCH", "HTTPS_STR_MATCH"]:
        string_match = module.params.get("string_match")
        if not string_match:
            missing_args.append("string_match")
        health_check["SearchString"] = module.params.get("string_match")

    if type_in == "CALCULATED":
        if not child_health_checks_in:
            missing_args.append("child_health_checks")
        if not health_threshold_in:
            missing_args.append("health_threshold")
        health_check["ChildHealthChecks"] = child_health_checks_in
        health_check["HealthThreshold"] = health_threshold_in
    else:
        failure_threshold = module.params.get("failure_threshold")
        if not failure_threshold:
            failure_threshold = 3
        health_check["FailureThreshold"] = failure_threshold
        health_check["RequestInterval"] = request_interval_in

    if module.params.get("measure_latency") is not None:
        health_check["MeasureLatency"] = module.params.get("measure_latency")

    if missing_args:
        module.fail_json(
            msg=f"missing required arguments for creation: {', '.join(missing_args)}",
        )

    if module.check_mode:
        return True, "create", None

    try:
        result = client.create_health_check(
            aws_retry=True,
            CallerReference=caller_ref,
            HealthCheckConfig=health_check,
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg="Failed to create health check.", health_check=health_check)

    check_id = result.get("HealthCheck").get("Id")
    return True, "create", check_id


def update_health_check(existing_check):
    # It's possible to update following parameters
    # - ResourcePath
    # - SearchString
    # - FailureThreshold
    # - Disabled
    # - IPAddress
    # - Port
    # - FullyQualifiedDomainName
    # - ChildHealthChecks
    # - HealthThreshold

    changes = dict()
    existing_config = existing_check.get("HealthCheckConfig")
    check_id = existing_check.get("Id")

    resource_path = module.params.get("resource_path", None)
    if resource_path and resource_path != existing_config.get("ResourcePath"):
        changes["ResourcePath"] = resource_path

    search_string = module.params.get("string_match", None)
    if search_string and search_string != existing_config.get("SearchString"):
        changes["SearchString"] = search_string

    type_in = module.params.get("type", None)
    if type_in != "CALCULATED":
        failure_threshold = module.params.get("failure_threshold", None)
        if failure_threshold and failure_threshold != existing_config.get("FailureThreshold"):
            changes["FailureThreshold"] = failure_threshold

    disabled = module.params.get("disabled", None)
    if disabled is not None and disabled != existing_config.get("Disabled"):
        changes["Disabled"] = module.params.get("disabled")

    # If updating based on Health Check ID or health_check_name, we can update
    if module.params.get("health_check_id") or module.params.get("use_unique_names"):
        ip_address = module.params.get("ip_address", None)
        if ip_address is not None and ip_address != existing_config.get("IPAddress"):
            changes["IPAddress"] = module.params.get("ip_address")

        port = module.params.get("port", None)
        if port is not None and port != existing_config.get("Port"):
            changes["Port"] = module.params.get("port")

        fqdn = module.params.get("fqdn", None)
        if fqdn is not None and fqdn != existing_config.get("FullyQualifiedDomainName"):
            changes["FullyQualifiedDomainName"] = module.params.get("fqdn")

        if type_in == "CALCULATED":
            child_health_checks = module.params.get("child_health_checks", None)
            if child_health_checks is not None and child_health_checks != existing_config.get("ChildHealthChecks"):
                changes["ChildHealthChecks"] = module.params.get("child_health_checks")

            health_threshold = module.params.get("health_threshold", None)
            if health_threshold is not None and health_threshold != existing_config.get("HealthThreshold"):
                changes["HealthThreshold"] = module.params.get("health_threshold")

    # No changes...
    if not changes:
        return False, None, check_id
    if module.check_mode:
        return True, "update", check_id

    # This makes sure we're starting from the version we think we are...
    version_id = existing_check.get("HealthCheckVersion", 1)
    try:
        client.update_health_check(
            HealthCheckId=check_id,
            HealthCheckVersion=version_id,
            **changes,
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg="Failed to update health check.", id=check_id)

    return True, "update", check_id


def describe_health_check(id):
    if not id:
        return dict()

    try:
        result = client.get_health_check(
            aws_retry=True,
            HealthCheckId=id,
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg="Failed to get health check.", id=id)

    health_check = result.get("HealthCheck", {})
    health_check = camel_dict_to_snake_dict(health_check)
    tags = get_tags(module, client, "healthcheck", id)
    health_check["tags"] = tags
    return health_check


def main():
    argument_spec = dict(
        state=dict(choices=["present", "absent"], default="present"),
        disabled=dict(type="bool"),
        ip_address=dict(),
        port=dict(type="int"),
        type=dict(choices=["HTTP", "HTTPS", "HTTP_STR_MATCH", "HTTPS_STR_MATCH", "TCP", "CALCULATED"]),
        child_health_checks=dict(type="list", elements="str"),
        health_threshold=dict(type="int", default=1),
        resource_path=dict(),
        fqdn=dict(),
        string_match=dict(),
        request_interval=dict(type="int", choices=[10, 30], default=30),
        failure_threshold=dict(type="int", choices=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(type="bool", default=True),
        health_check_id=dict(type="str", aliases=["id"], required=False),
        health_check_name=dict(type="str", aliases=["name"], required=False),
        use_unique_names=dict(type="bool", required=False),
        measure_latency=dict(type="bool", required=False),
    )

    args_one_of = [
        ["ip_address", "fqdn", "health_check_id", "child_health_checks"],
    ]

    args_if = [
        ["type", "TCP", ("port",)],
        ["type", "CALCULATED", ("child_health_checks", "health_threshold")],
    ]

    args_required_together = [
        ["use_unique_names", "health_check_name"],
    ]

    args_mutually_exclusive = [
        ["health_check_id", "health_check_name"],
        ["child_health_checks", "ip_address"],
        ["child_health_checks", "port"],
        ["child_health_checks", "fqdn"],
    ]

    global module
    global client

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        required_one_of=args_one_of,
        required_if=args_if,
        required_together=args_required_together,
        mutually_exclusive=args_mutually_exclusive,
        supports_check_mode=True,
    )

    if not module.params.get("health_check_id") and not module.params.get("type"):
        module.fail_json(msg="parameter 'type' is required if not updating or deleting health check by ID.")

    state_in = module.params.get("state")
    ip_addr_in = module.params.get("ip_address")
    port_in = module.params.get("port")
    type_in = module.params.get("type")
    fqdn_in = module.params.get("fqdn")
    string_match_in = module.params.get("string_match")
    request_interval_in = module.params.get("request_interval")
    health_check_name = module.params.get("health_check_name")
    tags = module.params.get("tags")
    purge_tags = module.params.get("purge_tags")
    child_health_checks_in = module.params.get("child_health_checks")
    health_threshold_in = module.params.get("health_threshold")

    # Default port
    if port_in is None:
        if type_in in ["HTTP", "HTTP_STR_MATCH"]:
            port_in = 80
        elif type_in in ["HTTPS", "HTTPS_STR_MATCH"]:
            port_in = 443

    if string_match_in:
        if type_in not in ["HTTP_STR_MATCH", "HTTPS_STR_MATCH"]:
            module.fail_json(msg="parameter 'string_match' argument is only for the HTTP(S)_STR_MATCH types")
        if len(string_match_in) > 255:
            module.fail_json(msg="parameter 'string_match' is limited to 255 characters max")

    client = module.client("route53", retry_decorator=AWSRetry.jittered_backoff())

    changed = False
    action = None
    check_id = None

    # If update or delete Health Check based on ID
    update_delete_by_id = False
    if module.params.get("health_check_id"):
        update_delete_by_id = True
        id_to_update_delete = module.params.get("health_check_id")
        try:
            existing_check = client.get_health_check(HealthCheckId=id_to_update_delete)["HealthCheck"]
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.exit_json(
                changed=False, msg=f"The specified health check with ID: {id_to_update_delete} does not exist"
            )
    else:
        existing_check = find_health_check(ip_addr_in, fqdn_in, type_in, request_interval_in, port_in)
        if existing_check:
            check_id = existing_check.get("Id")

    # Delete Health Check
    if state_in == "absent":
        if update_delete_by_id:
            changed, action = delete_health_check(id_to_update_delete)
        else:
            changed, action = delete_health_check(check_id)
        check_id = None

    # Create Health Check
    elif state_in == "present":
        if existing_check is None and not module.params.get("use_unique_names") and not update_delete_by_id:
            changed, action, check_id = create_health_check(
                ip_addr_in, fqdn_in, type_in, request_interval_in, port_in, child_health_checks_in, health_threshold_in
            )

        # Update Health Check
        else:
            # If health_check_name is a unique identifier
            if module.params.get("use_unique_names"):
                existing_checks_with_name = get_existing_checks_with_name()
                if tags is None:
                    purge_tags = False
                    tags = {}
                tags["Name"] = health_check_name

                # update the health_check if another health check with same name exists
                if health_check_name in existing_checks_with_name:
                    changed, action, check_id = update_health_check(existing_checks_with_name[health_check_name])
                else:
                    # create a new health_check if another health check with same name does not exists
                    changed, action, check_id = create_health_check(
                        ip_addr_in,
                        fqdn_in,
                        type_in,
                        request_interval_in,
                        port_in,
                        child_health_checks_in,
                        health_threshold_in,
                    )

            else:
                changed, action, check_id = update_health_check(existing_check)

        if check_id:
            changed |= manage_tags(module, client, "healthcheck", check_id, tags, purge_tags)

    health_check = describe_health_check(id=check_id)
    health_check["action"] = action
    module.exit_json(
        changed=changed,
        health_check=health_check,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
module: nxos_command
extends_documentation_fragment:
- cisco.nxos.nxos
author: Peter Sprygada (@privateip)
notes:
- Limited Support for Cisco MDS
short_description: Run arbitrary command on Cisco NXOS devices
description:
- Sends an arbitrary command to an NXOS node and returns the results read from the
  device.  This module includes an argument that will cause the module to wait for
  a specific condition before returning or timing out if the condition is not met.
version_added: 1.0.0
options:
  commands:
    description:
    - The commands to send to the remote NXOS device.  The resulting output from the
      command is returned.  If the I(wait_for) argument is provided, the module is
      not returned until the condition is satisfied or the number of retires as expired.
    - The I(commands) argument also accepts an alternative form that allows for complex
      values that specify the command to run and the output format to return. This
      can be done on a command by command basis.  The complex argument supports the
      keywords C(command) and C(output) where C(command) is the command to run and
      C(output) is one of 'text' or 'json'.
    - If a command sent to the device requires answering a prompt, it is possible to pass
      a dict containing command, answer and prompt. Common answers are 'y' or "\\r"
      (carriage return, must be double quotes). See examples.
    required: true
    type: list
    elements: raw
  wait_for:
    description:
    - Specifies what to evaluate from the output of the command and what conditionals
      to apply.  This argument will cause the task to wait for a particular conditional
      to be true before moving forward.   If the conditional is not true by the configured
      retries, the task fails.  See examples.
    aliases:
    - waitfor
    type: list
    elements: str
  match:
    description:
    - The I(match) argument is used in conjunction with the I(wait_for) argument to
      specify the match policy.  Valid values are C(all) or C(any).  If the value
      is set to C(all) then all conditionals in the I(wait_for) must be satisfied.  If
      the value is set to C(any) then only one of the values must be satisfied.
    default: all
    choices: ['any', 'all']
    type: str
  retries:
    description:
    - Specifies the number of retries a command should by tried before it is considered
      failed. The command is run on the target device every retry and evaluated against
      the I(wait_for) conditionals.
    - The commands are run once when I(retries) is set to C(0).
    default: 9
    type: int
  interval:
    description:
    - Configures the interval in seconds to wait between retries of the command.  If
      the command does not pass the specified conditional, the interval indicates
      how to long to wait before trying the command again.
    default: 1
    type: int
"""

EXAMPLES = """
- name: run show version on remote devices
  cisco.nxos.nxos_command:
    commands: show version

- name: run show version and check to see if output contains Cisco
  cisco.nxos.nxos_command:
    commands: show version
    wait_for: result[0] contains Cisco

- name: run multiple commands on remote nodes
  cisco.nxos.nxos_command:
    commands:
    - show version
    - show interfaces

- name: run multiple commands and evaluate the output
  cisco.nxos.nxos_command:
    commands:
    - show version
    - show interfaces
    wait_for:
    - result[0] contains Cisco
    - result[1] contains loopback0

- name: run commands and specify the output format
  cisco.nxos.nxos_command:
    commands:
    - command: show version
      output: json

- name: run commands that require answering a prompt
  cisco.nxos.nxos_command:
    commands:
    - configure terminal
    - command: no feature npv
      prompt: Do you want to continue
      answer: y

"""

RETURN = """
stdout:
  description: The set of responses from the commands
  returned: always apart from low level errors (such as action plugin)
  type: list
  sample: ['...', '...']
stdout_lines:
  description: The value of stdout split into a list
  returned: always apart from low level errors (such as action plugin)
  type: list
  sample: [['...', '...'], ['...'], ['...']]
failed_conditions:
  description: The list of conditionals that have failed
  returned: failed
  type: list
  sample: ['...', '...']
"""
import time

from ansible.module_utils._text import to_text
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.parsing import (
    Conditional,
    FailedConditionalError,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    to_lines,
    transform_commands,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import run_commands


def parse_commands(module, warnings):
    commands = transform_commands(module)

    if module.check_mode:
        for item in list(commands):
            if not item["command"].startswith("show"):
                warnings.append(
                    "Only show commands are supported when using check mode, not "
                    "executing %s" % item["command"],
                )
                commands.remove(item)

    return commands


def to_cli(obj):
    cmd = obj["command"]
    if obj.get("output") == "json":
        cmd += " | json"
    return cmd


def main():
    """entry point for module execution"""
    argument_spec = dict(
        # { command: <str>, output: <str>, prompt: <str>, response: <str> }
        commands=dict(type="list", required=True, elements="raw"),
        wait_for=dict(type="list", aliases=["waitfor"], elements="str"),
        match=dict(default="all", choices=["any", "all"]),
        retries=dict(default=9, type="int"),
        interval=dict(default=1, type="int"),
    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    warnings = list()
    result = {"changed": False, "warnings": warnings}
    commands = parse_commands(module, warnings)
    wait_for = module.params["wait_for"] or list()
    conditionals = []

    try:
        conditionals = [Conditional(c) for c in wait_for]
    except AttributeError as exc:
        module.fail_json(msg=to_text(exc))

    retries = module.params["retries"]
    interval = module.params["interval"]
    match = module.params["match"]

    while retries >= 0:
        responses = run_commands(module, commands)

        for item in list(conditionals):
            try:
                if item(responses):
                    if match == "any":
                        conditionals = list()
                        break
                    conditionals.remove(item)
            except FailedConditionalError as exc:
                module.fail_json(msg=to_text(exc))

        if not conditionals:
            break

        time.sleep(interval)
        retries -= 1

    if conditionals:
        failed_conditions = [item.raw for item in conditionals]
        msg = "One or more conditional statements have not been satisfied"
        module.fail_json(msg=msg, failed_conditions=failed_conditions)

    result.update({"stdout": responses, "stdout_lines": list(to_lines(responses))})

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2024, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type
__author__ = ("Madhan Sankaranarayanan, Rishita Chowdhary, Abhishek Maheshwari")

DOCUMENTATION = r"""
---
module: swim_workflow_manager
short_description: workflow_manager module for SWIM related functions
description:
- Manage operation related to image importation, distribution, activation and tagging image as golden
- API to fetch a software image from remote file system using URL for HTTP/FTP and upload it to Catalyst Center.
  Supported image files extensions are bin, img, tar, smu, pie, aes, iso, ova, tar_gz and qcow2.
- API to fetch a software image from local file system and upload it to Catalyst Center
  Supported image files extensions are bin, img, tar, smu, pie, aes, iso, ova, tar_gz and qcow2.
- API to tag/untag image as golen for a given family of devices
- API to distribute a software image on a given device. Software image must be imported successfully into
  Catalyst Center before it can be distributed.
- API to activate a software image on a given device. Software image must be present in the device flash.
version_added: '6.6.0'
extends_documentation_fragment:
  - cisco.dnac.workflow_manager_params
author: Madhan Sankaranarayanan (@madhansansel)
        Rishita Chowdhary (@rishitachowdhary)
        Abhishek Maheshwari (@abmahesh)
options:
  config_verify:
    description: Set to True to verify the Cisco Catalyst Center config after applying the playbook config.
    type: bool
    default: False
  state:
    description: The state of Catalyst Center after module completion.
    type: str
    choices: [ merged ]
    default: merged
  config:
    description: List of details of SWIM image being managed
    type: list
    elements: dict
    required: True
    suboptions:
      import_image_details:
        description: Details of image being imported
        type: dict
        suboptions:
          type:
            description: Specifies the import source, supporting local file import (local) or remote url import (remote).
            type: str
          local_image_details:
            description: Details of the local path of the image to be imported.
            type: dict
            suboptions:
              file_path:
                description: Provide the absolute file path needed to import an image from your local system (Eg "/path/to/your/file").
                    Accepted files formats are - .gz,.bin,.img,.tar,.smu,.pie,.aes,.iso,.ova,.tar_gz,.qcow2,.nfvispkg,.zip,.spa,.rpm.
                type: str
              is_third_party:
                description: Query parameter to determine if the image is from a third party (optional).
                type: bool
              third_party_application_type:
                description: Specify the ThirdPartyApplicationType query parameter to indicate the type of third-party application. Allowed
                    values include WLC, LINUX, FIREWALL, WINDOWS, LOADBALANCER, THIRDPARTY, etc.(optional).
                  WLC (Wireless LAN Controller) - It's a network device that manages and controls multiple wireless access points (APs) in a
                    centralized manner.
                  LINUX - It's an open-source operating system that provides a complete set of software packages and utilities.
                  FIREWALL - It's a network security device that monitors and controls incoming and outgoing network traffic based on
                    predetermined security rules.It acts as a barrier between a trusted internal network and untrusted external networks
                    (such as the internet), preventing unauthorized access.
                  WINDOWS - It's an operating system known for its graphical user interface (GUI) support, extensive compatibility with hardware
                    and software, and widespread use across various applications.
                  LOADBALANCER - It's a network device or software application that distributes incoming network traffic across multiple servers
                    or resources.
                  THIRDPARTY - It refers to third-party images or applications that are not part of the core system.
                  NAM (Network Access Manager) - It's a network management tool or software application that provides centralized control and
                    monitoring of network access policies, user authentication, and device compliance.
                  WAN Optimization - It refers to techniques and technologies used to improve the performance and efficiency of WANs. It includes
                    various optimization techniques such as data compression, caching, protocol optimization, and traffic prioritization to reduce
                    latency, increase throughput, and improve user experience over WAN connections.
                  Unknown - It refers to an unspecified or unrecognized application type.
                  Router - It's a network device that forwards data packets between computer networks. They are essential for connecting multiple
                    networks together and directing traffic between them.
                type: str
              third_party_image_family:
                description: Provide the ThirdPartyImageFamily query parameter to identify the family of the third-party image. Image Family name
                    like PALOALTO, RIVERBED, FORTINET, CHECKPOINT, SILVERPEAK etc. (optional).
                type: str
              third_party_vendor:
                description: Include the ThirdPartyVendor query parameter to specify the vendor of the third party.
                type: str
          url_details:
            description: URL details for SWIM import
            type: dict
            suboptions:
              payload:
                description: Swim Import Via Url's payload.
                type: list
                elements: dict
                suboptions:
                  application_type:
                    description: An optional parameter that specifies the type of application. Allowed values include WLC, LINUX, FIREWALL, WINDOWS,
                        LOADBALANCER, THIRDPARTY, etc. This is only applicable for third-party image types(optional).
                      WLC (Wireless LAN Controller) - It's network device that manages and controls multiple wireless access points (APs) in a
                        centralized manner.
                      LINUX - It's an open source which provide complete operating system with a wide range of software packages and utilities.
                      FIREWALL - It's a network security device that monitors and controls incoming and outgoing network traffic based on
                        predetermined security rules.It acts as a barrier between a trusted internal network and untrusted external networks
                        (such as the internet), preventing unauthorized access.
                      WINDOWS - It's an OS which provides GUI support for various applications, and extensive compatibility with hardware
                        and software.
                      LOADBALANCER - It's a network device or software application that distributes incoming network traffic across multiple servers
                        or resources.
                      THIRDPARTY - It refers to third-party images or applications that are not part of the core system.
                      NAM (Network Access Manager) - It's a network management tool or software application that provides centralized control and
                        monitoring of network access policies, user authentication, and device compliance.
                      WAN Optimization - It refers to techniques and technologies used to improve the performance and efficiency of WANs. It includes
                        various optimization techniques such as data compression, caching, protocol optimization, and traffic prioritization to reduce
                        latency, increase throughput, and improve user experience over WAN connections.
                      Unknown - It refers to an unspecified or unrecognized application type.
                      Router - It's a network device that forwards data packets between computer networks. They are essential for connecting multiple
                        networks together and directing traffic between them.
                    type: str
                  image_family:
                    description: Represents the name of the image family and is applicable only when uploading third-party images. Image Family name
                        like PALOALTO, RIVERBED, FORTINET, CHECKPOINT, SILVERPEAK etc. (optional).
                    type: str
                  source_url:
                    description: A mandatory parameter for importing a SWIM image via a remote URL. This parameter is required when using a URL
                        to import an image..(For example, http://{host}/swim/cat9k_isoxe.16.12.10s.SPA.bin,
                        ftp://user:password@{host}/swim/cat9k_isoxe.16.12.10s.SPA.iso)
                    type: str
                  is_third_party:
                    description: Flag indicates whether the image is uploaded from a third party (optional).
                    type: bool
                  vendor:
                    description: The name of the vendor, that applies only to third-party image types when importing via URL (optional).
                    type: str
              schedule_at:
                description: ScheduleAt query parameter. Epoch Time (The number of milli-seconds since
                  January 1 1970 UTC) at which the distribution should be scheduled (optional).
                type: str
              schedule_desc:
                description: ScheduleDesc query parameter. Custom Description (optional).
                type: str
              schedule_origin:
                description: ScheduleOrigin query parameter. Originator of this call (optional).
                type: str
      tagging_details:
        description: Details for tagging or untagging an image as golden
        type: dict
        suboptions:
          image_name:
            description: SWIM image name which will be tagged or untagged as golden.
            type: str
          device_role:
            description: Defines the device role, with permissible values including ALL, UNKNOWN, ACCESS, BORDER ROUTER,
              DISTRIBUTION, and CORE.
              ALL - This role typically represents all devices within the network, regardless of their specific roles or functions.
              UNKNOWN - This role is assigned to devices whose roles or functions have not been identified or classified within Cisco Catalsyt Center.
                This could happen if the platform is unable to determine the device's role based on available information.
              ACCESS - This role typically represents switches or access points that serve as access points for end-user devices to connect to the network.
                These devices are often located at the edge of the network and provide connectivity to end-user devices.
              BORDER ROUTER - These are devices that connect different network domains or segments together. They often serve as
                gateways between different networks, such as connecting an enterprise network to the internet or connecting
                multiple branch offices.
              DISTRIBUTION - This role represents function as distribution switches or routers in hierarchical network designs. They aggregate traffic
                from access switches and route it toward the core of the network or toward other distribution switches.
              CORE - This role typically represents high-capacity switches or routers that form the backbone of the network. They handle large volumes
                of traffic and provide connectivity between different parts of network, such as connecting distribution switches or
                providing interconnection between different network segments.
            type: str
          device_image_family_name:
            description: Device Image family name(Eg Cisco Catalyst 9300 Switch)
            type: str
          site_name:
            description: Site name for which SWIM image will be tagged/untagged as golden.
              If not provided, SWIM image will be mapped to global site.
            type: str
          tagging:
            description: Booelan value to tag/untag SWIM image as golden
              If True then the given image will be tagged as golden.
              If False then the given image will be un-tagged as golden.
            type: bool
      image_distribution_details:
        description: Details for SWIM image distribution. Device on which the image needs to distributed
          can be speciifed using any of the following parameters - deviceSerialNumber,
          deviceIPAddress, deviceHostname or deviceMacAddress.
        type: dict
        suboptions:
          device_role:
            description: Device Role and  permissible Values are ALL, UNKNOWN, ACCESS, BORDER ROUTER,
              DISTRIBUTION and CORE.
              ALL - This role typically represents all devices within the network, regardless of their specific roles or functions.
              UNKNOWN - This role is assigned to devices whose roles or functions have not been identified or classified within Cisco Catalsyt Center.
                This could happen if the platform is unable to determine the device's role based on available information.
              ACCESS - This role typically represents switches or access points that serve as access points for end-user devices to connect to the network.
                These devices are often located at the edge of the network and provide connectivity to end-user devices.
              BORDER ROUTER - These are devices that connect different network domains or segments together. They often serve as
                gateways between different networks, such as connecting an enterprise network to the internet or connecting
                multiple branch offices.
              DISTRIBUTION - This role represents function as distribution switches or routers in hierarchical network designs. They aggregate traffic
                from access switches and route it toward the core of the network or toward other distribution switches.
              CORE - This role typically represents high-capacity switches or routers that form the backbone of the network. They handle large volumes
                of traffic and provide connectivity between different parts of network, such as connecting distribution switches or
                providing interconnection between different network segments.
            type: str
          device_family_name:
            description: Specify the name of the device family such as Switches and Hubs, etc.
            type: str
          site_name:
            description: Used to get device details associated to this site.
            type: str
          device_series_name:
            description: This parameter specifies the name of the device series. It is used to identify a specific series of devices,
                such as Cisco Catalyst 9300 Series Switches, within the Cisco Catalyst Center.
            type: str
            version_added: 6.12.0
          image_name:
            description: SWIM image's name
            type: str
          device_serial_number:
            description: Device serial number where the image needs to be distributed
            type: str
          device_ip_address:
            description: Device IP address where the image needs to be distributed
            type: str
          device_hostname:
            description: Device hostname where the image needs to be distributed
            type: str
          device_mac_address:
            description: Device MAC address where the image needs to be distributed
            type: str
      image_activation_details:
        description: Details for SWIM image activation. Device on which the image needs to activated
          can be speciifed using any of the following parameters - deviceSerialNumber,
          deviceIPAddress, deviceHostname or deviceMacAddress.
        type: dict
        suboptions:
          device_role:
            description: Defines the device role, with permissible values including ALL, UNKNOWN, ACCESS, BORDER ROUTER,
              DISTRIBUTION, and CORE.
            type: str
          device_family_name:
            description: Specify the name of the device family such as Switches and Hubs, etc.
            type: str
          site_name:
            description: Used to get device details associated to this site.
            type: str
          device_series_name:
            description: This parameter specifies the name of the device series. It is used to identify a specific series of devices,
                such as Cisco Catalyst 9300 Series Switches, within the Cisco Catalyst Center.
            type: str
            version_added: 6.12.0
          activate_lower_image_version:
            description: ActivateLowerImageVersion flag.
            type: bool
          device_upgrade_mode:
            description: It specifies the mode of upgrade to be applied to the devices having the following values - 'install', 'bundle', and 'currentlyExists'.
              install - This mode instructs Cisco Catalyst Center to perform a clean installation of the new image on the target devices.
                When this mode is selected, the existing image on the device is completely replaced with the new image during the upgrade process.
                This ensures that the device runs only the new image version after the upgrade is completed.
              bundle - This mode instructs Cisco Catalyst Center bundles the new image with the existing image on the device before initiating
                the upgrade process. This mode allows for a more efficient upgrade process by preserving the existing image on the device while
                adding the new image as an additional bundle. After the upgrade, the device can run either the existing image or the new bundled
                image, depending on the configuration.
              currentlyExists - This mode instructs Cisco Catalyst Center to checks if the target devices already have the desired image version
                installed. If image already present on devices, no action is taken and upgrade process is skipped for those devices. This mode
                is useful for avoiding unnecessary upgrades on devices that already have the correct image version installed, thereby saving time.
            type: str
          distribute_if_needed:
            description: Enable the distribute_if_needed option when activating the SWIM image.
            type: bool
          image_name:
            description: SWIM image's name
            type: str
          device_serial_number:
            description: Device serial number where the image needs to be activated
            type: str
          device_ip_address:
            description: Device IP address where the image needs to be activated
            type: str
          device_hostname:
            description: Device hostname where the image needs to be activated
            type: str
          device_mac_address:
            description: Device MAC address where the image needs to be activated
            type: str
          schedule_validate:
            description: ScheduleValidate query parameter. ScheduleValidate, validates data
              before schedule (optional).
            type: bool
requirements:
- dnacentersdk == 2.4.5
- python >= 3.5
notes:
  - SDK Method used are
    software_image_management_swim.SoftwareImageManagementSwim.import_software_image_via_url,
    software_image_management_swim.SoftwareImageManagementSwim.tag_as_golden_image,
    software_image_management_swim.SoftwareImageManagementSwim.trigger_software_image_distribution,
    software_image_management_swim.SoftwareImageManagementSwim.trigger_software_image_activation,

  - Paths used are
    post /dna/intent/api/v1/image/importation/source/url,
    post /dna/intent/api/v1/image/importation/golden,
    post /dna/intent/api/v1/image/distribution,
    post /dna/intent/api/v1/image/activation/device,

  - Added the parameter 'dnac_api_task_timeout', 'dnac_task_poll_interval' options in v6.13.2.

"""

EXAMPLES = r"""
- name: Import an image from a URL, tag it as golden and load it on device
  cisco.dnac.swim_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log_level: "{{dnac_log_level}}"
    dnac_log: True
    config:
    - import_image_details:
        type: remote
        url_details:
          payload:
          - source_url: "http://10.10.10.10/stda/cat9k_iosxe.17.12.01.SPA.bin"
            is_third_party: False
      tagging_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        device_role: ACCESS
        device_image_family_name: Cisco Catalyst 9300 Switch
        site_name: Global/USA/San Francisco/BGL_18
        tagging: True
      image_distribution_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        device_serial_number: FJC2327U0S2
      image_activation_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        schedule_validate: False
        activate_lower_image_version: False
        distribute_if_needed: True
        device_serial_number: FJC2327U0S2

- name: Import an image from local, tag it as golden.
  cisco.dnac.swim_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log_level: "{{dnac_log_level}}"
    dnac_log: True
    config:
    - import_image_details:
        type: local
        local_image_details:
            file_path: /Users/Downloads/cat9k_iosxe.17.12.01.SPA.bin
            is_third_party: False
      tagging_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        device_role: ACCESS
        device_image_family_name: Cisco Catalyst 9300 Switch
        site_name: Global/USA/San Francisco/BGL_18
        tagging: True

- name: Tag the given image as golden and load it on device
  cisco.dnac.swim_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log_level: "{{dnac_log_level}}"
    dnac_log: True
    config:
    - tagging_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        device_role: ACCESS
        device_image_family_name: Cisco Catalyst 9300 Switch
        site_name: Global/USA/San Francisco/BGL_18
        tagging: True

- name: Un-tagged the given image as golden and load it on device
  cisco.dnac.swim_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log_level: "{{dnac_log_level}}"
    dnac_log: True
    config:
    - tagging_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        device_role: ACCESS
        device_image_family_name: Cisco Catalyst 9300 Switch
        site_name: Global/USA/San Francisco/BGL_18
        tagging: False

- name: Distribute the given image on devices associated to that site with specified role.
  cisco.dnac.swim_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log_level: "{{dnac_log_level}}"
    dnac_log: True
    config:
    - image_distribution_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        site_name: Global/USA/San Francisco/BGL_18
        device_role: ALL
        device_family_name: Switches and Hubs
        device_series_name: Cisco Catalyst 9300 Series Switches

- name: Activate the given image on devices associated to that site with specified role.
  cisco.dnac.swim_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log_level: "{{dnac_log_level}}"
    dnac_log: True
    config:
    - image_activation_details:
        image_name: cat9k_iosxe.17.12.01.SPA.bin
        site_name: Global/USA/San Francisco/BGL_18
        device_role: ALL
        device_family_name: Switches and Hubs
        device_series_name: Cisco Catalyst 9300 Series Switches
        scehdule_validate: False
        activate_lower_image_version: True
        distribute_if_needed: True

"""

RETURN = r"""
#Case: SWIM image is successfully imported, tagged as golden, distributed and activated on a device
response:
  description: A dictionary with activation details as returned by the Catalyst Center Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
                        "additionalStatusURL": String,
                        "data": String,
                        "endTime": 0,
                        "id": String,
                        "instanceTenantId": String,
                        "isError": bool,
                        "lastUpdate": 0,
                        "progress": String,
                        "rootId": String,
                        "serviceType": String,
                        "startTime": 0,
                        "version": 0
                  },
      "msg": String
    }

"""

from ansible_collections.cisco.dnac.plugins.module_utils.dnac import (
    DnacBase,
    validate_list_of_dicts,
    get_dict_result,
)
from ansible.module_utils.basic import AnsibleModule
import os
import time


class Swim(DnacBase):
    """Class containing member attributes for Swim workflow_manager module"""

    def __init__(self, module):
        super().__init__(module)
        self.supported_states = ["merged"]

    def validate_input(self):
        """
        Validate the fields provided in the playbook.
        Checks the configuration provided in the playbook against a predefined specification
        to ensure it adheres to the expected structure and data types.
        Parameters:
          - self: The instance of the class containing the 'config' attribute to be validated.
        Returns:
          The method returns an instance of the class with updated attributes:
          - self.msg: A message describing the validation result.
          - self.status: The status of the validation (either 'success' or 'failed').
          - self.validated_config: If successful, a validated version of 'config' parameter.
        Example:
            To use this method, create an instance of the class and call 'validate_input' on it.
          If the validation succeeds, 'self.status' will be 'success' and 'self.validated_config'
          will contain the validated configuration. If it fails, 'self.status' will be 'failed',
          'self.msg' will describe the validation issues.
        """

        if not self.config:
            self.status = "success"
            self.msg = "Configuration is not available in the playbook for validation"
            self.log(self.msg, "ERROR")
            return self

        temp_spec = dict(
            import_image_details=dict(type='dict'),
            tagging_details=dict(type='dict'),
            image_distribution_details=dict(type='dict'),
            image_activation_details=dict(type='dict'),
        )

        # Validate swim params
        valid_temp, invalid_params = validate_list_of_dicts(
            self.config, temp_spec
        )

        if invalid_params:
            self.msg = "Invalid parameters in playbook: {0}".format(invalid_params)
            self.log(self.msg, "ERROR")
            self.status = "failed"
            return self

        self.validated_config = valid_temp
        self.msg = "Successfully validated playbook config params: {0}".format(str(valid_temp))
        self.log(self.msg, "INFO")
        self.status = "success"

        return self

    def site_exists(self, site_name):
        """
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            tuple: A tuple containing two values:
            - site_exists (bool): A boolean indicating whether the site exists (True) or not (False).
            - site_id (str or None): The ID of the site if it exists, or None if the site is not found.
        Description:
            This method checks the existence of a site in the Catalyst Center. If the site is found,it sets 'site_exists' to True,
            retrieves the site's ID, and returns both values in a tuple. If the site does not exist, 'site_exists' is set
            to False, and 'site_id' is None. If an exception occurs during the site lookup, an exception is raised.
        """

        site_exists = False
        site_id = None
        response = None
        try:
            response = self.dnac._exec(
                family="sites",
                function='get_site',
                params={"name": site_name},
            )
        except Exception as e:
            self.msg = "An exception occurred: Site '{0}' does not exist in the Cisco Catalyst Center".format(site_name)
            self.log(self.msg, "ERROR")
            self.module.fail_json(msg=self.msg)

        if response:
            self.log("Received API response from 'get_site': {0}".format(str(response)), "DEBUG")
            site = response.get("response")
            site_id = site[0].get("id")
            site_exists = True

        return (site_exists, site_id)

    def get_image_id(self, name):
        """
        Retrieve the unique image ID based on the provided image name.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            name (str): The name of the software image to search for.
        Returns:
            str: The unique image ID (UUID) corresponding to the given image name.
        Raises:
            AnsibleFailJson: If the image is not found in the response.
        Description:
            This function sends a request to Cisco Catalyst Center to retrieve details about a software image based on its name.
            It extracts and returns the image ID if a single matching image is found. If no image or multiple
            images are found with the same name, it raises an exception.
        """

        image_response = self.dnac._exec(
            family="software_image_management_swim",
            function='get_software_image_details',
            params={"image_name": name},
        )
        self.log("Received API response from 'get_software_image_details': {0}".format(str(image_response)), "DEBUG")
        image_list = image_response.get("response")

        if (len(image_list) == 1):
            image_id = image_list[0].get("imageUuid")
            self.log("SWIM image '{0}' has the ID: {1}".format(name, image_id), "INFO")
        else:
            error_message = "SWIM image '{0}' could not be found".format(name)
            self.log(error_message, "ERROR")
            self.module.fail_json(msg=error_message, response=image_response)

        return image_id

    def get_image_name_from_id(self, image_id):
        """
        Retrieve the unique image name based on the provided image id.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            id (str): The unique image ID (UUID) of the software image to search for.
        Returns:
            str: The image name corresponding to the given unique image ID (UUID)
        Raises:
            AnsibleFailJson: If the image is not found in the response.
        Description:
            This function sends a request to Cisco Catalyst Center to retrieve details about a software image based on its id.
            It extracts and returns the image name if a single matching image is found. If no image or multiple
            images are found with the same name, it raises an exception.
        """

        image_response = self.dnac._exec(
            family="software_image_management_swim",
            function='get_software_image_details',
            params={"image_uuid": image_id},
        )
        self.log("Received API response from 'get_software_image_details': {0}".format(str(image_response)), "DEBUG")
        image_list = image_response.get("response")

        if (len(image_list) == 1):
            image_name = image_list[0].get("name")
            self.log("SWIM image '{0}' has been fetched successfully from Cisco Catalyst Center".format(image_name), "INFO")
        else:
            error_message = "SWIM image with Id '{0}' could not be found in Cisco Catalyst Center".format(image_id)
            self.log(error_message, "ERROR")
            self.module.fail_json(msg=error_message, response=image_response)

        return image_name

    def is_image_exist(self, name):
        """
        Retrieve the unique image ID based on the provided image name.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            name (str): The name of the software image to search for.
        Returns:
            str: The unique image ID (UUID) corresponding to the given image name.
        Raises:
            AnsibleFailJson: If the image is not found in the response.
        Description:
            This function sends a request to Cisco Catalyst Center to retrieve details about a software image based on its name.
            It extracts and returns the image ID if a single matching image is found. If no image or multiple
            images are found with the same name, it raises an exception.
        """

        image_exist = False
        image_response = self.dnac._exec(
            family="software_image_management_swim",
            function='get_software_image_details',
            params={"image_name": name},
        )
        self.log("Received API response from 'get_software_image_details': {0}".format(str(image_response)), "DEBUG")
        image_list = image_response.get("response")

        if (len(image_list) == 1):
            image_exist = True

        return image_exist

    def get_device_id(self, params):
        """
        Retrieve the unique device ID based on the provided parameters.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            params (dict): A dictionary containing parameters to filter devices.
        Returns:
            str: The unique device ID corresponding to the filtered device.
        Description:
            This function sends a request to Cisco Catalyst Center to retrieve a list of devices based on the provided
            filtering parameters. If a single matching device is found, it extracts and returns the device ID. If
            no device or multiple devices match the criteria, it raises an exception.
        """
        device_id = None
        response = self.dnac._exec(
            family="devices",
            function='get_device_list',
            params=params,
        )
        self.log("Received API response from 'get_device_list': {0}".format(str(response)), "DEBUG")

        device_list = response.get("response")
        if (len(device_list) == 1):
            device_id = device_list[0].get("id")
            self.log("Device Id: {0}".format(str(device_id)), "INFO")
        else:
            self.msg = "Device with params: '{0}' not found in Cisco Catalyst Center so can't fetch the device id".format(str(params))
            self.log(self.msg, "WARNING")

        return device_id

    def get_device_uuids(self, site_name, device_family, device_role, device_series_name=None):
        """
        Retrieve a list of device UUIDs based on the specified criteria.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            site_name (str): The name of the site for which device UUIDs are requested.
            device_family (str): The family/type of devices to filter on.
            device_role (str): The role of devices to filter on. If None, 'ALL' roles are considered.
            device_series_name(str): Specifies the name of the device series.
        Returns:
            list: A list of device UUIDs that match the specified criteria.
        Description:
            The function checks the reachability status and role of devices in the given site.
            Only devices with "Reachable" status are considered, and filtering is based on the specified
            device family and role (if provided).
        """

        device_uuid_list = []
        if not site_name:
            site_name = "Global"
            self.log("Site name not specified; defaulting to 'Global' to fetch all devices under this category", "INFO")

        (site_exists, site_id) = self.site_exists(site_name)
        if not site_exists:
            self.log("""Site '{0}' is not found in the Cisco Catalyst Center, hence unable to fetch associated
                        devices.""".format(site_name), "INFO")
            return device_uuid_list

        if device_series_name:
            if device_series_name.startswith(".*") and device_series_name.endswith(".*"):
                self.log("Device series name '{0}' is already in the regex format".format(device_series_name), "INFO")
            else:
                device_series_name = ".*" + device_series_name + ".*"

        site_params = {
            "site_id": site_id,
            "device_family": device_family
        }
        response = self.dnac._exec(
            family="sites",
            function='get_membership',
            op_modifies=True,
            params=site_params,
        )
        self.log("Received API response from 'get_membership': {0}".format(str(response)), "DEBUG")
        response = response['device']

        site_response_list = []
        for item in response:
            if item['response']:
                for item_dict in item['response']:
                    site_response_list.append(item_dict)

        if device_role.upper() == 'ALL':
            device_role = None

        device_params = {
            'series': device_series_name,
            'family': device_family,
            'role': device_role
        }
        device_list_response = self.dnac._exec(
            family="devices",
            function='get_device_list',
            op_modifies=True,
            params=device_params,
        )

        device_response = device_list_response.get('response')
        if not response or not device_response:
            self.log("Failed to retrieve devices associated with the site '{0}' due to empty API response.".format(site_name), "INFO")
            return device_uuid_list

        site_memberships_ids, device_response_ids = [], []

        for item in site_response_list:
            if item["reachabilityStatus"] != "Reachable":
                self.log("""Device '{0}' is currently '{1}' and cannot be included in the SWIM distribution/activation
                            process.""".format(item["managementIpAddress"], item["reachabilityStatus"]), "INFO")
                continue
            self.log("""Device '{0}' from site '{1}' is ready for the SWIM distribution/activation
                        process.""".format(item["managementIpAddress"], site_name), "INFO")
            site_memberships_ids.append(item["instanceUuid"])

        for item in device_response:
            if item["reachabilityStatus"] != "Reachable":
                self.log("""Unable to proceed with the device '{0}' for SWIM distribution/activation as its status is
                            '{1}'.""".format(item["managementIpAddress"], item["reachabilityStatus"]), "INFO")
                continue
            self.log("""Device '{0}' matches to the specified filter requirements and is set for SWIM
                      distribution/activation.""".format(item["managementIpAddress"]), "INFO")
            device_response_ids.append(item["instanceUuid"])

        # Find the intersection of device IDs with the response get from get_membership api and get_device_list api with provided filters
        device_uuid_list = set(site_memberships_ids).intersection(set(device_response_ids))

        return device_uuid_list

    def get_device_family_identifier(self, family_name):
        """
        Retrieve and store the device family identifier based on the provided family name.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            family_name (str): The name of the device family for which to retrieve the identifier.
        Returns:
            None
        Raises:
            AnsibleFailJson: If the family name is not found in the response.
        Description:
            This function sends a request to Cisco Catalyst Center to retrieve a list of device family identifiers.It then
            searches for a specific family name within the response and stores its associated identifier. If the family
            name is found, the identifier is stored; otherwise, an exception is raised.
        """

        have = {}
        response = self.dnac._exec(
            family="software_image_management_swim",
            function='get_device_family_identifiers',
        )
        self.log("Received API response from 'get_device_family_identifiers': {0}".format(str(response)), "DEBUG")
        device_family_db = response.get("response")

        if device_family_db:
            device_family_details = get_dict_result(device_family_db, 'deviceFamily', family_name)

            if device_family_details:
                device_family_identifier = device_family_details.get("deviceFamilyIdentifier")
                have["device_family_identifier"] = device_family_identifier
                self.log("Family device indentifier: {0}".format(str(device_family_identifier)), "INFO")
            else:
                self.msg = "Device Family: {0} not found".format(str(family_name))
                self.log(self.msg, "ERROR")
                self.module.fail_json(msg=self.msg, response=self.msg)
            self.have.update(have)

    def get_have(self):
        """
        Retrieve and store various software image and device details based on user-provided information.
        Returns:
            self: The current instance of the class with updated 'have' attributes.
        Raises:
            AnsibleFailJson: If required image or device details are not provided.
        Description:
            This function populates the 'have' dictionary with details related to software images, site information,
            device families, distribution devices, and activation devices based on user-provided data in the 'want' dictionary.
            It validates and retrieves the necessary information from Cisco Catalyst Center to support later actions.
        """

        if self.want.get("tagging_details"):
            have = {}
            tagging_details = self.want.get("tagging_details")
            if tagging_details.get("image_name"):
                name = tagging_details.get("image_name").split("/")[-1]
                image_id = self.get_image_id(name)
                have["tagging_image_id"] = image_id

            elif self.have.get("imported_image_id"):
                have["tagging_image_id"] = self.have.get("imported_image_id")

            else:
                self.log("Image details for tagging not provided", "CRITICAL")
                self.module.fail_json(msg="Image details for tagging not provided", response=[])

            # check if given site exists, store siteid
            # if not then use global site
            site_name = tagging_details.get("site_name")
            if site_name:
                site_exists = False
                (site_exists, site_id) = self.site_exists(site_name)
                if site_exists:
                    have["site_id"] = site_id
                    self.log("Site {0} exists having the site id: {1}".format(site_name, str(site_id)), "DEBUG")
            else:
                # For global site, use -1 as siteId
                have["site_id"] = "-1"
                self.log("Site Name not given by user. Using global site.", "WARNING")

            self.have.update(have)
            # check if given device family name exists, store indentifier value
            family_name = tagging_details.get("device_image_family_name")
            self.get_device_family_identifier(family_name)

        if self.want.get("distribution_details"):
            have = {}
            distribution_details = self.want.get("distribution_details")
            site_name = distribution_details.get("site_name")
            if site_name:
                site_exists = False
                (site_exists, site_id) = self.site_exists(site_name)

                if site_exists:
                    have["site_id"] = site_id
                    self.log("Site '{0}' exists and has the site ID: {1}".format(site_name, str(site_id)), "DEBUG")

            # check if image for distributon is available
            if distribution_details.get("image_name"):
                name = distribution_details.get("image_name").split("/")[-1]
                image_id = self.get_image_id(name)
                have["distribution_image_id"] = image_id

            elif self.have.get("imported_image_id"):
                have["distribution_image_id"] = self.have.get("imported_image_id")

            else:
                self.log("Image details required for distribution have not been provided", "ERROR")
                self.module.fail_json(msg="Image details required for distribution have not been provided", response=[])

            device_params = dict(
                hostname=distribution_details.get("device_hostname"),
                serialNumber=distribution_details.get("device_serial_number"),
                managementIpAddress=distribution_details.get("device_ip_address"),
                macAddress=distribution_details.get("device_mac_address"),
            )
            device_id = self.get_device_id(device_params)

            if device_id is not None:
                have["distribution_device_id"] = device_id

            self.have.update(have)

        if self.want.get("activation_details"):
            have = {}
            activation_details = self.want.get("activation_details")
            # check if image for activation is available
            if activation_details.get("image_name"):
                name = activation_details.get("image_name").split("/")[-1]
                image_id = self.get_image_id(name)
                have["activation_image_id"] = image_id

            elif self.have.get("imported_image_id"):
                have["activation_image_id"] = self.have.get("imported_image_id")
            else:
                self.log("Image details required for activation have not been provided", "ERROR")
                self.module.fail_json(msg="Image details required for activation have not been provided", response=[])

            site_name = activation_details.get("site_name")
            if site_name:
                site_exists = False
                (site_exists, site_id) = self.site_exists(site_name)
                if site_exists:
                    have["site_id"] = site_id
                    self.log("The site '{0}' exists and has the site ID '{1}'".format(site_name, str(site_id)), "INFO")

            device_params = dict(
                hostname=activation_details.get("device_hostname"),
                serialNumber=activation_details.get("device_serial_number"),
                managementIpAddress=activation_details.get("device_ip_address"),
                macAddress=activation_details.get("device_mac_address"),
            )
            device_id = self.get_device_id(device_params)

            if device_id is not None:
                have["activation_device_id"] = device_id
            self.have.update(have)
            self.log("Current State (have): {0}".format(str(self.have)), "INFO")

        return self

    def get_want(self, config):
        """
        Retrieve and store import, tagging, distribution, and activation details from playbook configuration.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            config (dict): The configuration dictionary containing image import and other details.
        Returns:
            self: The current instance of the class with updated 'want' attributes.
        Raises:
            AnsibleFailJson: If an incorrect import type is specified.
        Description:
            This function parses the playbook configuration to extract information related to image
            import, tagging, distribution, and activation. It stores these details in the 'want' dictionary
            for later use in the Ansible module.
        """

        want = {}
        if config.get("import_image_details"):
            want["import_image"] = True
            want["import_type"] = config.get("import_image_details").get("type").lower()
            if want["import_type"] == "remote":
                want["url_import_details"] = config.get("import_image_details").get("url_details")
            elif want["import_type"] == "local":
                want["local_import_details"] = config.get("import_image_details").get("local_image_details")
            else:
                self.log("The import type '{0}' provided is incorrect. Only 'local' or 'remote' are supported.".format(want["import_type"]), "CRITICAL")
                self.module.fail_json(msg="Incorrect import type. Supported Values: local or remote")

        want["tagging_details"] = config.get("tagging_details")
        want["distribution_details"] = config.get("image_distribution_details")
        want["activation_details"] = config.get("image_activation_details")

        self.want = want
        self.log("Desired State (want): {0}".format(str(self.want)), "INFO")

        return self

    def get_diff_import(self):
        """
        Check the image import type and fetch the image ID for the imported image for further use.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This function checks the type of image import (URL or local) and proceeds with the import operation accordingly.
            It then monitors the import task's progress and updates the 'result' dictionary. If the operation is successful,
            'changed' is set to True.
            Additionally, if tagging, distribution, or activation details are provided, it fetches the image ID for the
            imported image and stores it in the 'have' dictionary for later use.
        """

        try:
            import_type = self.want.get("import_type")

            if not import_type:
                self.status = "success"
                self.msg = "Error: Details required for importing SWIM image. Please provide the necessary information."
                self.result['msg'] = self.msg
                self.log(self.msg, "WARNING")
                self.result['changed'] = False
                return self

            if import_type == "remote":
                image_name = self.want.get("url_import_details").get("payload")[0].get("source_url")
            else:
                image_name = self.want.get("local_import_details").get("file_path")

            # Code to check if the image already exists in Catalyst Center
            name = image_name.split('/')[-1]
            image_exist = self.is_image_exist(name)

            import_key_mapping = {
                'source_url': 'sourceURL',
                'image_family': 'imageFamily',
                'application_type': 'applicationType',
                'is_third_party': 'thirdParty',
            }

            if image_exist:
                image_id = self.get_image_id(name)
                self.have["imported_image_id"] = image_id
                self.msg = "Image '{0}' already exists in the Cisco Catalyst Center".format(name)
                self.result['msg'] = self.msg
                self.log(self.msg, "INFO")
                self.status = "success"
                self.result['changed'] = False
                return self

            if self.want.get("import_type") == "remote":
                import_payload_dict = {}
                temp_payload = self.want.get("url_import_details").get("payload")[0]
                keys_to_change = list(import_key_mapping.keys())

                for key, val in temp_payload.items():
                    if key in keys_to_change:
                        api_key_name = import_key_mapping[key]
                        import_payload_dict[api_key_name] = val

                import_image_payload = [import_payload_dict]
                import_params = dict(
                    payload=import_image_payload,
                    scheduleAt=self.want.get("url_import_details").get("schedule_at"),
                    scheduleDesc=self.want.get("url_import_details").get("schedule_desc"),
                    scheduleOrigin=self.want.get("url_import_details").get("schedule_origin"),
                )
                import_function = 'import_software_image_via_url'
            else:
                file_path = self.want.get("local_import_details").get("file_path")
                import_params = dict(
                    is_third_party=self.want.get("local_import_details").get("is_third_party"),
                    third_party_vendor=self.want.get("local_import_details").get("third_party_vendor"),
                    third_party_image_family=self.want.get("local_import_details").get("third_party_image_family"),
                    third_party_application_type=self.want.get("local_import_details").get("third_party_application_type"),
                    multipart_fields={'file': (os.path.basename(file_path), open(file_path, 'rb'), 'application/octet-stream')},
                    multipart_monitor_callback=None
                )
                import_function = 'import_local_software_image'

            response = self.dnac._exec(
                family="software_image_management_swim",
                function=import_function,
                op_modifies=True,
                params=import_params,
            )
            self.log("Received API response from {0}: {1}".format(import_function, str(response)), "DEBUG")

            task_details = {}
            task_id = response.get("response").get("taskId")

            while (True):
                task_details = self.get_task_details(task_id)
                name = image_name.split('/')[-1]

                if task_details and \
                        ("completed successfully" in task_details.get("progress").lower()):
                    self.result['changed'] = True
                    self.status = "success"
                    self.msg = "Swim Image {0} imported successfully".format(name)
                    self.result['msg'] = self.msg
                    self.log(self.msg, "INFO")
                    break

                if task_details and task_details.get("isError"):
                    if "already exists" in task_details.get("failureReason", ""):
                        self.msg = "SWIM Image {0} already exists in the Cisco Catalyst Center".format(name)
                        self.result['msg'] = self.msg
                        self.log(self.msg, "INFO")
                        self.status = "success"
                        self.result['changed'] = False
                        break
                    else:
                        self.status = "failed"
                        self.msg = task_details.get("failureReason", "SWIM Image {0} seems to be invalid".format(image_name))
                        self.log(self.msg, "WARNING")
                        self.result['response'] = self.msg
                        return self

            self.result['response'] = task_details if task_details else response

            # Fetch image_id for the imported image for further use
            image_name = image_name.split('/')[-1]
            image_id = self.get_image_id(image_name)
            self.have["imported_image_id"] = image_id

            return self

        except Exception as e:
            self.status = "failed"
            self.msg = """Error: Import image details are not provided in the playbook, or the Import Image API was not
                 triggered successfully. Please ensure the necessary details are provided and verify the status of the Import Image process."""
            self.log(self.msg, "ERROR")
            self.result['response'] = self.msg

        return self

    def get_diff_tagging(self):
        """
        Tag or untag a software image as golden based on provided tagging details.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This function tags or untags a software image as a golden image in Cisco Catalyst Center based on the provided
            tagging details. The tagging action is determined by the value of the 'tagging' attribute
            in the 'tagging_details' dictionary.If 'tagging' is True, the image is tagged as golden, and if 'tagging'
            is False, the golden tag is removed. The function sends the appropriate request to Cisco Catalyst Center and updates the
            task details in the 'result' dictionary. If the operation is successful, 'changed' is set to True.
        """

        tagging_details = self.want.get("tagging_details")
        tag_image_golden = tagging_details.get("tagging")
        image_name = self.get_image_name_from_id(self.have.get("tagging_image_id"))

        image_params = dict(
            image_id=self.have.get("tagging_image_id"),
            site_id=self.have.get("site_id"),
            device_family_identifier=self.have.get("device_family_identifier"),
            device_role=tagging_details.get("device_role", "ALL").upper()
        )

        response = self.dnac._exec(
            family="software_image_management_swim",
            function='get_golden_tag_status_of_an_image',
            op_modifies=True,
            params=image_params
        )
        self.log("Received API response from 'get_golden_tag_status_of_an_image': {0}".format(str(response)), "DEBUG")

        response = response.get('response')
        if response:
            image_status = response['taggedGolden']
            if image_status and image_status == tag_image_golden:
                self.status = "success"
                self.result['changed'] = False
                self.msg = "SWIM Image '{0}' already tagged as Golden image in Cisco Catalyst Center".format(image_name)
                self.result['msg'] = self.msg
                self.log(self.msg, "INFO")
                return self

            if not image_status and image_status == tag_image_golden:
                self.status = "success"
                self.result['changed'] = False
                self.msg = "SWIM Image '{0}' already un-tagged from Golden image in Cisco Catalyst Center".format(image_name)
                self.result['msg'] = self.msg
                self.log(self.msg, "INFO")
                return self

        if tag_image_golden:
            image_params = dict(
                imageId=self.have.get("tagging_image_id"),
                siteId=self.have.get("site_id"),
                deviceFamilyIdentifier=self.have.get("device_family_identifier"),
                deviceRole=tagging_details.get("device_role", "ALL").upper()
            )
            self.log("Parameters for tagging the image as golden: {0}".format(str(image_params)), "INFO")

            response = self.dnac._exec(
                family="software_image_management_swim",
                function='tag_as_golden_image',
                op_modifies=True,
                params=image_params
            )
            self.log("Received API response from 'tag_as_golden_image': {0}".format(str(response)), "DEBUG")

        else:
            self.log("Parameters for un-tagging the image as golden: {0}".format(str(image_params)), "INFO")

            response = self.dnac._exec(
                family="software_image_management_swim",
                function='remove_golden_tag_for_image',
                op_modifies=True,
                params=image_params
            )
            self.log("Received API response from 'remove_golden_tag_for_image': {0}".format(str(response)), "DEBUG")

        if not response:
            self.status = "failed"
            self.msg = "Did not get the response of API so cannot check the Golden tagging status of image - {0}".format(image_name)
            self.log(self.msg, "ERROR")
            self.result['response'] = self.msg
            return self

        task_details = {}
        task_id = response.get("response").get("taskId")

        while True:
            task_details = self.get_task_details(task_id)

            if not task_details.get("isError") and 'successful' in task_details.get("progress"):
                self.status = "success"
                self.result['changed'] = True
                self.msg = task_details.get("progress")
                self.result['msg'] = self.msg
                self.result['response'] = self.msg
                self.log(self.msg, "INFO")
                break
            elif task_details.get("isError"):
                failure_reason = task_details.get("failureReason", "")
                if failure_reason and "An inheritted tag cannot be un-tagged" in failure_reason:
                    self.status = "failed"
                    self.result['changed'] = False
                    self.msg = failure_reason
                    self.result['msg'] = failure_reason
                    self.log(self.msg, "ERROR")
                    self.result['response'] = self.msg
                    break
                else:
                    error_message = task_details.get("failureReason", "Error: while tagging/un-tagging the golden swim image.")
                    self.status = "failed"
                    self.msg = error_message
                    self.result['msg'] = error_message
                    self.log(self.msg, "ERROR")
                    self.result['response'] = self.msg
                    break

        return self

    def get_device_ip_from_id(self, device_id):
        """
        Retrieve the management IP address of a device from Cisco Catalyst Center using its ID.
        Parameters:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            - device_id (str): The unique identifier of the device in Cisco Catalyst Center.
        Returns:
            str: The management IP address of the specified device.
        Raises:
            Exception: If there is an error while retrieving the response from Cisco Catalyst Center.
        Description:
            This method queries Cisco Catalyst Center for the device details based on its unique identifier (ID).
            It uses the 'get_device_list' function in the 'devices' family, extracts the management IP address
            from the response, and returns it. If any error occurs during the process, an exception is raised
            with an appropriate error message logged.
        """

        try:
            response = self.dnac._exec(
                family="devices",
                function='get_device_list',
                params={"id": device_id}
            )
            self.log("Received API response from 'get_device_list': {0}".format(str(response)), "DEBUG")
            response = response.get('response')[0]
            device_ip = response.get("managementIpAddress")

            return device_ip
        except Exception as e:
            error_message = "Error occurred while getting the response of device from Cisco Catalyst Center: {0}".format(str(e))
            self.log(error_message, "ERROR")
            raise Exception(error_message)

    def check_swim_task_status(self, swim_task_dict, swim_task_name):
        """
        Check the status of the SWIM (Software Image Management) task for each device.
        Args:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            swim_task_dict (dict): A dictionary containing the mapping of device IP address to the respective task ID.
            swim_task_name (str): The name of the SWIM task being checked which is either Distribution or Activation.
        Returns:
            tuple: A tuple containing two elements:
                - device_ips_list (list): A list of device IP addresses for which the SWIM task failed.
                - device_count (int): The count of devices for which the SWIM task was successful.
        Description:
            This function iterates through the distribution_task_dict, which contains the mapping of
            device IP address to their respective task ID. It checks the status of the SWIM task for each device by
            repeatedly querying for task details until the task is either completed successfully or fails. If the task
            is successful, the device count is incremented. If the task fails, an error message is logged, and the device
            IP is appended to the device_ips_list and return a tuple containing the device_ips_list and device_count.
        """

        device_ips_list = []
        device_count = 0

        for device_ip, task_id in swim_task_dict.items():
            start_time = time.time()

            while (True):
                end_time = time.time()
                max_timeout = self.params.get('dnac_api_task_timeout')

                if (end_time - start_time) >= max_timeout:
                    self.log("""Max timeout of {0} has reached for the task id '{1}' for the device '{2}' and unexpected
                                 task status so moving out to next task id""".format(max_timeout, task_id, device_ip), "WARNING")
                    device_ips_list.append(device_ip)
                    break

                task_details = self.get_task_details(task_id)

                if not task_details.get("isError") and \
                        ("completed successfully" in task_details.get("progress")):
                    self.result['changed'] = True
                    self.status = "success"
                    self.log("Image {0} successfully for the device '{1}".format(swim_task_name, device_ip), "INFO")
                    device_count += 1
                    break

                if task_details.get("isError"):
                    error_msg = "Image {0} gets failed for the device '{1}'".format(swim_task_name, device_ip)
                    self.log(error_msg, "ERROR")
                    self.result['response'] = task_details
                    device_ips_list.append(device_ip)
                    break
                time.sleep(self.params.get('dnac_task_poll_interval'))

        return device_ips_list, device_count

    def get_diff_distribution(self):
        """
        Get image distribution parameters from the playbook and trigger image distribution.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This function retrieves image distribution parameters from the playbook's 'distribution_details' and triggers
            the distribution of the specified software image to the specified device. It monitors the distribution task's
            progress and updates the 'result' dictionary. If the operation is successful, 'changed' is set to True.
        """

        distribution_details = self.want.get("distribution_details")
        site_name = distribution_details.get("site_name")
        device_family = distribution_details.get("device_family_name")
        device_role = distribution_details.get("device_role", "ALL")
        device_series_name = distribution_details.get("device_series_name")
        device_uuid_list = self.get_device_uuids(site_name, device_family, device_role, device_series_name)
        image_id = self.have.get("distribution_image_id")
        self.complete_successful_distribution = False
        self.partial_successful_distribution = False
        self.single_device_distribution = False

        if self.have.get("distribution_device_id"):
            distribution_params = dict(
                payload=[dict(
                    deviceUuid=self.have.get("distribution_device_id"),
                    imageUuid=image_id
                )]
            )
            self.log("Distribution Params: {0}".format(str(distribution_params)), "INFO")

            response = self.dnac._exec(
                family="software_image_management_swim",
                function='trigger_software_image_distribution',
                op_modifies=True,
                params=distribution_params,
            )
            self.log("Received API response from 'trigger_software_image_distribution': {0}".format(str(response)), "DEBUG")

            if response:
                task_details = {}
                task_id = response.get("response").get("taskId")

                while (True):
                    task_details = self.get_task_details(task_id)

                    if not task_details.get("isError") and \
                            ("completed successfully" in task_details.get("progress")):
                        self.result['changed'] = True
                        self.status = "success"
                        self.single_device_distribution = True
                        self.result['msg'] = "Image with Id {0} Distributed Successfully".format(image_id)
                        break

                    if task_details.get("isError"):
                        self.status = "failed"
                        self.msg = "Image with Id {0} Distribution Failed".format(image_id)
                        self.log(self.msg, "ERROR")
                        self.result['response'] = task_details
                        break

                    self.result['response'] = task_details if task_details else response

            return self

        if len(device_uuid_list) == 0:
            self.status = "success"
            self.msg = "The SWIM image distribution task could not proceed because no eligible devices were found"
            self.result['msg'] = self.msg
            self.log(self.msg, "WARNING")
            return self

        self.log("Device UUIDs involved in Image Distribution: {0}".format(str(device_uuid_list)), "INFO")
        distribution_task_dict = {}

        for device_uuid in device_uuid_list:
            device_management_ip = self.get_device_ip_from_id(device_uuid)
            distribution_params = dict(
                payload=[dict(
                    deviceUuid=device_uuid,
                    imageUuid=image_id
                )]
            )
            self.log("Distribution Params: {0}".format(str(distribution_params)), "INFO")
            response = self.dnac._exec(
                family="software_image_management_swim",
                function='trigger_software_image_distribution',
                op_modifies=True,
                params=distribution_params,
            )
            self.log("Received API response from 'trigger_software_image_distribution': {0}".format(str(response)), "DEBUG")

            if response:
                task_details = {}
                task_id = response.get("response").get("taskId")
                distribution_task_dict[device_management_ip] = task_id

        device_ips_list, device_distribution_count = self.check_swim_task_status(distribution_task_dict, 'Distribution')

        if device_distribution_count == 0:
            self.status = "failed"
            self.msg = "Image with Id {0} Distribution Failed for all devices".format(image_id)
        elif device_distribution_count == len(device_uuid_list):
            self.result['changed'] = True
            self.status = "success"
            self.complete_successful_distribution = True
            self.msg = "Image with Id {0} Distributed Successfully for all devices".format(image_id)
        else:
            self.result['changed'] = True
            self.status = "success"
            self.partial_successful_distribution = False
            self.msg = "Image with Id '{0}' Distributed and partially successfull".format(image_id)
            self.log("For device(s) {0} image Distribution gets failed".format(str(device_ips_list)), "CRITICAL")

        self.result['msg'] = self.msg
        self.log(self.msg, "INFO")

        return self

    def get_diff_activation(self):
        """
        Get image activation parameters from the playbook and trigger image activation.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This function retrieves image activation parameters from the playbook's 'activation_details' and triggers the
            activation of the specified software image on the specified device. It monitors the activation task's progress and
            updates the 'result' dictionary. If the operation is successful, 'changed' is set to True.
        """

        activation_details = self.want.get("activation_details")
        site_name = activation_details.get("site_name")
        device_family = activation_details.get("device_family_name")
        device_role = activation_details.get("device_role", "ALL")
        device_series_name = activation_details.get("device_series_name")
        device_uuid_list = self.get_device_uuids(site_name, device_family, device_role, device_series_name)
        image_id = self.have.get("activation_image_id")
        self.complete_successful_activation = False
        self.partial_successful_activation = False
        self.single_device_activation = False

        if self.have.get("activation_device_id"):
            payload = [dict(
                activateLowerImageVersion=activation_details.get("activate_lower_image_version"),
                deviceUpgradeMode=activation_details.get("device_upgrade_mode"),
                distributeIfNeeded=activation_details.get("distribute_if_needed"),
                deviceUuid=self.have.get("activation_device_id"),
                imageUuidList=[image_id]
            )]

            activation_params = dict(
                schedule_validate=activation_details.get("scehdule_validate"),
                payload=payload
            )
            self.log("Activation Params: {0}".format(str(activation_params)), "INFO")

            response = self.dnac._exec(
                family="software_image_management_swim",
                function='trigger_software_image_activation',
                op_modifies=True,
                params=activation_params,
            )
            self.log("Received API response from 'trigger_software_image_activation': {0}".format(str(response)), "DEBUG")

            task_details = {}
            task_id = response.get("response").get("taskId")

            while (True):
                task_details = self.get_task_details(task_id)

                if not task_details.get("isError") and \
                        ("completed successfully" in task_details.get("progress")):
                    self.result['changed'] = True
                    self.result['msg'] = "Image Activated successfully"
                    self.status = "success"
                    self.single_device_activation = True
                    break

                if task_details.get("isError"):
                    self.msg = "Activation for Image with Id '{0}' gets failed".format(image_id)
                    self.status = "failed"
                    self.result['response'] = task_details
                    self.log(self.msg, "ERROR")
                    return self

            self.result['response'] = task_details if task_details else response

            return self

        if len(device_uuid_list) == 0:
            self.status = "success"
            self.msg = "The SWIM image activation task could not proceed because no eligible devices were found."
            self.result['msg'] = self.msg
            self.log(self.msg, "WARNING")
            return self

        self.log("Device UUIDs involved in Image Activation: {0}".format(str(device_uuid_list)), "INFO")
        activation_task_dict = {}

        for device_uuid in device_uuid_list:
            device_management_ip = self.get_device_ip_from_id(device_uuid)
            payload = [dict(
                activateLowerImageVersion=activation_details.get("activate_lower_image_version"),
                deviceUpgradeMode=activation_details.get("device_upgrade_mode"),
                distributeIfNeeded=activation_details.get("distribute_if_needed"),
                deviceUuid=device_uuid,
                imageUuidList=[image_id]
            )]

            activation_params = dict(
                schedule_validate=activation_details.get("scehdule_validate"),
                payload=payload
            )
            self.log("Activation Params: {0}".format(str(activation_params)), "INFO")

            response = self.dnac._exec(
                family="software_image_management_swim",
                function='trigger_software_image_activation',
                op_modifies=True,
                params=activation_params,
            )
            self.log("Received API response from 'trigger_software_image_activation': {0}".format(str(response)), "DEBUG")

            if response:
                task_details = {}
                task_id = response.get("response").get("taskId")
                activation_task_dict[device_management_ip] = task_id

        device_ips_list, device_activation_count = self.check_swim_task_status(activation_task_dict, 'Activation')

        if device_activation_count == 0:
            self.status = "failed"
            self.msg = "Image with Id '{0}' activation failed for all devices".format(image_id)
        elif device_activation_count == len(device_uuid_list):
            self.result['changed'] = True
            self.status = "success"
            self.complete_successful_activation = True
            self.msg = "Image with Id '{0}' activated successfully for all devices".format(image_id)
        else:
            self.result['changed'] = True
            self.status = "success"
            self.partial_successful_activation = True
            self.msg = "Image with Id '{0}' activated and partially successful.".format(image_id)
            self.log("For Device(s) {0} Image activation gets Failed".format(str(device_ips_list)), "CRITICAL")

        self.result['msg'] = self.msg
        self.log(self.msg, "INFO")

        return self

    def get_diff_merged(self, config):
        """
        Get tagging details and then trigger distribution followed by activation if specified in the playbook.
        Parameters:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            config (dict): The configuration dictionary containing tagging, distribution, and activation details.
        Returns:
            self: The current instance of the class with updated 'result' and 'have' attributes.
        Description:
            This function checks the provided playbook configuration for tagging, distribution, and activation details. It
            then triggers these operations in sequence if the corresponding details are found in the configuration.The
            function monitors the progress of each task and updates the 'result' dictionary accordingly. If any of the
            operations are successful, 'changed' is set to True.
        """

        if config.get("tagging_details"):
            self.get_diff_tagging().check_return_status()

        if config.get("image_distribution_details"):
            self.get_diff_distribution().check_return_status()

        if config.get("image_activation_details"):
            self.get_diff_activation().check_return_status()

        return self

    def verify_diff_imported(self, import_type):
        """
        Verify the successful import of a software image into Cisco Catalyst Center.
        Args:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            import_type (str): The type of import, either 'remote' or 'local'.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method verifies the successful import of a software image into Cisco Catalyst Center.
            It checks whether the image exists in Catalyst Center based on the provided import type.
            If the image exists, the status is set to 'success', and a success message is logged.
            If the image does not exist, a warning message is logged indicating a potential import failure.
        """

        if import_type == "remote":
            image_name = self.want.get("url_import_details").get("payload")[0].get("source_url")
        else:
            image_name = self.want.get("local_import_details").get("file_path")

        # Code to check if the image already exists in Catalyst Center
        name = image_name.split('/')[-1]
        image_exist = self.is_image_exist(name)
        if image_exist:
            self.status = "success"
            self.msg = "The requested Image '{0}' imported in the Cisco Catalyst Center and Image presence has been verified.".format(name)
            self.log(self.msg, "INFO")
        else:
            self.log("""The playbook input for SWIM Image '{0}' does not align with the Cisco Catalyst Center, indicating that image
                        may not have imported successfully.""".format(name), "INFO")

        return self

    def verify_diff_tagged(self):
        """
        Verify the Golden tagging status of a software image in Cisco Catalyst Center.
        Args:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method verifies the tagging status of a software image in Cisco Catalyst Center.
            It retrieves tagging details from the input, including the desired tagging status and image ID.
            Using the provided image ID, it obtains image parameters required for checking the image status.
            The method then queries Catalyst Center to get the golden tag status of the image.
            If the image status matches the desired tagging status, a success message is logged.
            If there is a mismatch between the playbook input and the Catalyst Center, a warning message is logged.
        """

        tagging_details = self.want.get("tagging_details")
        tag_image_golden = tagging_details.get("tagging")
        image_id = self.have.get("tagging_image_id")
        image_name = self.get_image_name_from_id(image_id)

        image_params = dict(
            image_id=self.have.get("tagging_image_id"),
            site_id=self.have.get("site_id"),
            device_family_identifier=self.have.get("device_family_identifier"),
            device_role=tagging_details.get("device_role", "ALL").upper()
        )
        self.log("Parameters for checking the status of image: {0}".format(str(image_params)), "INFO")

        response = self.dnac._exec(
            family="software_image_management_swim",
            function='get_golden_tag_status_of_an_image',
            op_modifies=True,
            params=image_params
        )
        self.log("Received API response from 'get_golden_tag_status_of_an_image': {0}".format(str(response)), "DEBUG")

        response = response.get('response')
        if response:
            image_status = response['taggedGolden']
            if image_status == tag_image_golden:
                if tag_image_golden:
                    self.msg = """The requested image '{0}' has been tagged as golden in the Cisco Catalyst Center and
                             its status has been successfully verified.""".format(image_name)
                    self.log(self.msg, "INFO")
                else:
                    self.msg = """The requested image '{0}' has been un-tagged as golden in the Cisco Catalyst Center and
                            image status has been verified.""".format(image_name)
                    self.log(self.msg, "INFO")
        else:
            self.log("""Mismatch between the playbook input for tagging/un-tagging image as golden and the Cisco Catalyst Center indicates that
                        the tagging/un-tagging task was not executed successfully.""", "INFO")

        return self

    def verify_diff_distributed(self):
        """
        Verify the distribution status of a software image in Cisco Catalyst Center.
        Args:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method verifies the distribution status of a software image in Cisco Catalyst Center.
            It retrieves the image ID and name from the input and if distribution device ID is provided, it checks the distribution status for that
            list of specific device and logs the info message based on distribution status.
        """

        image_id = self.have.get("distribution_image_id")
        image_name = self.get_image_name_from_id(image_id)

        if self.have.get("distribution_device_id"):
            if self.single_device_distribution:
                self.msg = """The requested image '{0}', associated with the device ID '{1}', has been successfully distributed in the Cisco Catalyst Center
                     and its status has been verified.""".format(image_name, self.have.get("distribution_device_id"))
                self.log(self.msg, "INFO")
            else:
                self.log("""Mismatch between the playbook input for distributing the image to the device with ID '{0}' and the actual state in the
                         Cisco Catalyst Center suggests that the distribution task might not have been executed
                         successfully.""".format(self.have.get("distribution_device_id")), "INFO")
        elif self.complete_successful_distribution:
            self.msg = """The requested image '{0}', with ID '{1}', has been successfully distributed to all devices within the specified
                     site in the Cisco Catalyst Center.""".format(image_name, image_id)
            self.log(self.msg, "INFO")
        elif self.partial_successful_distribution:
            self.msg = """T"The requested image '{0}', with ID '{1}', has been partially distributed across some devices in the Cisco Catalyst
                     Center.""".format(image_name, image_id)
            self.log(self.msg, "INFO")
        else:
            self.msg = """The requested image '{0}', with ID '{1}', failed to be distributed across devices in the Cisco Catalyst
                     Center.""".format(image_name, image_id)
            self.log(self.msg, "INFO")

        return self

    def verify_diff_activated(self):
        """
        Verify the activation status of a software image in Cisco Catalyst Center.
        Args:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Returns:
            self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method verifies the activation status of a software image in Cisco Catalyst Center and retrieves the image ID and name from
            the input. If activation device ID is provided, it checks the activation status for that specific device. Based on activation status
            a corresponding message is logged.
        """

        image_id = self.have.get("activation_image_id")
        image_name = self.get_image_name_from_id(image_id)

        if self.have.get("activation_device_id"):
            if self.single_device_activation:
                self.msg = """The requested image '{0}', associated with the device ID '{1}', has been successfully activated in the Cisco Catalyst
                         Center and its status has been verified.""".format(image_name, self.have.get("activation_device_id"))
                self.log(self.msg, "INFO")
            else:
                self.log("""Mismatch between the playbook's input for activating the image '{0}' on the device with ID '{1}' and the actual state in
                         the Cisco Catalyst Center suggests that the activation task might not have been executed
                         successfully.""".format(image_name, self.have.get("activation_device_id")), "INFO")
        elif self.complete_successful_activation:
            self.msg = """The requested image '{0}', with ID '{1}', has been successfully activated on all devices within the specified site in the
                     Cisco Catalyst Center.""".format(image_name, image_id)
            self.log(self.msg, "INFO")
        elif self.partial_successful_activation:
            self.msg = """"The requested image '{0}', with ID '{1}', has been partially activated on some devices in the Cisco
                     Catalyst Center.""".format(image_name, image_id)
            self.log(self.msg, "INFO")
        else:
            self.msg = """The activation of the requested image '{0}', with ID '{1}', failed on devices in the Cisco
                     Catalyst Center.""".format(image_name, image_id)
            self.log(self.msg, "INFO")

        return self

    def verify_diff_merged(self, config):
        """
        Verify the merged status(Importing/Tagging/Distributing/Actiavting) the SWIM Image in devices in Cisco Catalyst Center.
        Args:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
            - config (dict): The configuration details to be verified.
        Return:
            - self (object): An instance of a class used for interacting with Cisco Catalyst Center.
        Description:
            This method checks the merged status of a configuration in Cisco Catalyst Center by retrieving the current state
            (have) and desired state (want) of the configuration, logs the states, and validates whether the specified
            SWIM operation performed or not.
        """

        self.get_have()
        self.log("Current State (have): {0}".format(str(self.have)), "INFO")
        self.log("Desired State (want): {0}".format(str(self.want)), "INFO")

        import_type = self.want.get("import_type")
        if import_type:
            self.verify_diff_imported(import_type).check_return_status()

        tagged = self.want.get("tagging_details")
        if tagged:
            self.verify_diff_tagged().check_return_status()

        distribution_details = self.want.get("distribution_details")
        if distribution_details:
            self.verify_diff_distributed().check_return_status()

        activation_details = self.want.get("activation_details")
        if activation_details:
            self.verify_diff_activated().check_return_status()

        return self


def main():
    """ main entry point for module execution
    """

    element_spec = {'dnac_host': {'required': True, 'type': 'str'},
                    'dnac_port': {'type': 'str', 'default': '443'},
                    'dnac_username': {'type': 'str', 'default': 'admin', 'aliases': ['user']},
                    'dnac_password': {'type': 'str', 'no_log': True},
                    'dnac_verify': {'type': 'bool', 'default': 'True'},
                    'dnac_version': {'type': 'str', 'default': '2.2.3.3'},
                    'dnac_debug': {'type': 'bool', 'default': False},
                    'dnac_log_level': {'type': 'str', 'default': 'WARNING'},
                    "dnac_log_file_path": {"type": 'str', "default": 'dnac.log'},
                    "dnac_log_append": {"type": 'bool', "default": True},
                    'dnac_log': {'type': 'bool', 'default': False},
                    'validate_response_schema': {'type': 'bool', 'default': True},
                    'config_verify': {'type': 'bool', "default": False},
                    'dnac_api_task_timeout': {'type': 'int', "default": 1200},
                    'dnac_task_poll_interval': {'type': 'int', "default": 2},
                    'config': {'required': True, 'type': 'list', 'elements': 'dict'},
                    'state': {'default': 'merged', 'choices': ['merged']}
                    }

    module = AnsibleModule(argument_spec=element_spec,
                           supports_check_mode=False)

    ccc_swims = Swim(module)
    state = ccc_swims.params.get("state")

    if state not in ccc_swims.supported_states:
        ccc_swims.status = "invalid"
        ccc_swims.msg = "State {0} is invalid".format(state)
        ccc_swims.check_return_status()

    ccc_swims.validate_input().check_return_status()
    config_verify = ccc_swims.params.get("config_verify")

    for config in ccc_swims.validated_config:
        ccc_swims.reset_values()
        ccc_swims.get_want(config).check_return_status()
        ccc_swims.get_diff_import().check_return_status()
        ccc_swims.get_have().check_return_status()
        ccc_swims.get_diff_state_apply[state](config).check_return_status()
        if config_verify:
            ccc_swims.verify_diff_state_apply[state](config).check_return_status()

    module.exit_json(**ccc_swims.result)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Akini Ross (@akinross) <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = """
    name: interface_range
    short_description: query interfaces from a range or comma separated list of ranges
    description:
      - this lookup returns interfaces from a range or comma separated list of ranges given to it
    notes:
      - duplicate interfaces from overlapping ranges will only be returned once
    options:
      _terms:
        description: comma separated strings of interface ranges
        required: True
"""

EXAMPLES = """
- name: "loop through range of interfaces"
  ansible.builtin.debug:
    msg: "{{ item }}"
  with_items: "{{ query('cisco.aci.interface_range', '1/1-4,1/20-25', '1/5', '1/2/3/8-10', '5/0-2') }}"
"""

RETURN = """
  _list:
    description: list of interfaces
    type: list
    elements: str
"""

import re

from ansible.errors import AnsibleError
from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    def run(self, terms, **kwargs):
        interfaces = []
        errors = []

        for interface_range in ",".join(terms).replace(" ", "").split(","):
            if re.fullmatch(r"((\d+/)+\d+-\d+$)", interface_range):
                slots = interface_range.rsplit("/", 1)[0]
                range_start, range_stop = interface_range.rsplit("/", 1)[1].split("-")
                if int(range_stop) > int(range_start):
                    for x in range(int(range_start), int(range_stop) + 1):
                        interfaces.append("{0}/{1}".format(slots, x))
                else:
                    errors.append(interface_range)
            elif re.fullmatch(r"((\d+/)+\d+$)", interface_range):
                interfaces.append(interface_range)
            else:
                errors.append(interface_range)
        if errors:
            raise AnsibleError("Invalid range inputs, {0}".format(errors))

        # Sorted functionality for visual aid only, will result in 1/25, 1/3, 1/31
        # If full sort is needed leverage natsort package (https://github.com/SethMMorton/natsort)
        return sorted(set(interfaces))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


import abc

from ansible.module_utils import six

from ansible.module_utils.common.validation import (
    check_type_str,
    check_type_list,
    check_type_dict,
    check_type_bool,
    check_type_int,
    check_type_float,
)


def ensure_type(value, type_name):
    if type_name == 'str':
        return check_type_str(value)
    if type_name == 'list':
        return check_type_list(value)
    if type_name == 'dict':
        return check_type_dict(value)
    if type_name == 'bool':
        return check_type_bool(value)
    if type_name == 'int':
        return check_type_int(value)
    if type_name == 'float':
        return check_type_float(value)
    return value


@six.add_metaclass(abc.ABCMeta)
class ProviderInformation(object):
    @abc.abstractmethod
    def get_zone_id_type(self):
        """
        Return the (short) type for zone IDs, like ``'int'`` or ``'str'``.
        """

    @abc.abstractmethod
    def get_record_id_type(self):
        """
        Return the (short) type for record IDs, like ``'int'`` or ``'str'``.
        """

    @abc.abstractmethod
    def get_record_default_ttl(self):
        """
        Return the default TTL for records, like 300, 3600 or None.
        None means that some other TTL (usually from the zone) will be used.
        """

    @abc.abstractmethod
    def get_supported_record_types(self):
        """
        Return a list of supported record types.
        """

    def normalize_prefix(self, prefix):
        """
        Given a prefix (string or None), return its normalized form.

        The result should always be None for the trivial prefix, and a non-zero length DNS name
        for a non-trivial prefix.

        If a provider supports other identifiers for the trivial prefix, such as '@', this
        function needs to convert them to None as well.
        """
        return prefix or None

    def supports_bulk_actions(self):
        """
        Return whether the API supports some kind of bulk actions.
        """
        return False

    @abc.abstractmethod
    def txt_record_handling(self):
        """
        Return how the API handles TXT records.

        Returns one of the following strings:
        * 'decoded' - the API works with unencoded values
        * 'encoded' - the API works with encoded values
        * 'encoded-no-char-encoding' - the API works with encoded values, but without character encoding
        """

    def txt_character_encoding(self):
        """
        Return how the API handles escape sequences in TXT records.

        Returns one of the following strings:
        * 'octal' - the API works with octal escape sequences
        * 'decimal' - the API works with decimal escape sequences

        This return value is only used if txt_record_handling returns 'encoded'.

        WARNING: the default return value will change to 'decimal' for community.dns 3.0.0!
        """
        return 'octal'



#!/usr/bin/python
#
# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: docker_image_tag

short_description: Tag Docker images with new names and/or tags

version_added: 3.6.0

description:
  - This module allows to tag Docker images with new names and/or tags.

extends_documentation_fragment:
  - community.docker.docker.api_documentation
  - community.docker.attributes
  - community.docker.attributes.actiongroup_docker

attributes:
  check_mode:
    support: full
  diff_mode:
    support: full

options:
  name:
    description:
      - "Image name. Name format will be one of: C(name), C(repository/name), C(registry_server:port/name).
        When pushing or pulling an image the name can optionally include the tag by appending C(:tag_name)."
      - Note that image IDs (hashes) can also be used.
    type: str
    required: true
  tag:
    description:
      - Tag for the image name O(name) that is to be tagged.
      - If O(name)'s format is C(name:tag), then the tag value from O(name) will take precedence.
    type: str
    default: latest
  repository:
    description:
      - List of new image names to tag the image as.
      - Expects format C(repository:tag). If no tag is provided, will use the value of the O(tag) parameter if present, or V(latest).
    type: list
    elements: str
    required: true
  existing_images:
    description:
      - Defines the behavior if the image to be tagged already exists and is another image than the one identified by O(name) and O(tag).
      - If set to V(keep), the tagged image is kept.
      - If set to V(overwrite), the tagged image is overwritten by the specified one.
    type: str
    choices:
      - keep
      - overwrite
    default: overwrite

requirements:
  - "Docker API >= 1.25"

author:
  - Felix Fontein (@felixfontein)

seealso:
  - module: community.docker.docker_image_push
  - module: community.docker.docker_image_remove
'''

EXAMPLES = '''
- name: Tag Python 3.12 image with two new names
  community.docker.docker_image_tag:
    name: python:3.12
    repository:
      - python-3:3.12
      - local-registry:5000/python-3/3.12:latest
'''

RETURN = '''
image:
    description: Image inspection results for the affected image.
    returned: success
    type: dict
    sample: {}
tagged_images:
    description:
      - A list of images that got tagged.
    returned: success
    type: list
    elements: str
    sample:
      - python-3:3.12
'''

import traceback

from ansible.module_utils.common.text.converters import to_native
from ansible.module_utils.common.text.formatters import human_to_bytes

from ansible_collections.community.docker.plugins.module_utils.common_api import (
    AnsibleDockerClient,
    RequestException,
)

from ansible_collections.community.docker.plugins.module_utils.util import (
    DockerBaseClass,
    is_image_name_id,
    is_valid_tag,
)

from ansible_collections.community.docker.plugins.module_utils._api.errors import DockerException
from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    parse_repository_tag,
)


def convert_to_bytes(value, module, name, unlimited_value=None):
    if value is None:
        return value
    try:
        if unlimited_value is not None and value in ('unlimited', str(unlimited_value)):
            return unlimited_value
        return human_to_bytes(value)
    except ValueError as exc:
        module.fail_json(msg='Failed to convert %s to bytes: %s' % (name, to_native(exc)))


def image_info(name, tag, image):
    result = dict(name=name, tag=tag)
    if image:
        result['id'] = image['Id']
    else:
        result['exists'] = False
    return result


class ImageTagger(DockerBaseClass):
    def __init__(self, client):
        super(ImageTagger, self).__init__()

        self.client = client
        parameters = self.client.module.params
        self.check_mode = self.client.check_mode

        self.name = parameters['name']
        self.tag = parameters['tag']
        if not is_valid_tag(self.tag, allow_empty=True):
            self.fail('"{0}" is not a valid docker tag'.format(self.tag))

        # If name contains a tag, it takes precedence over tag parameter.
        if not is_image_name_id(self.name):
            repo, repo_tag = parse_repository_tag(self.name)
            if repo_tag:
                self.name = repo
                self.tag = repo_tag

        self.keep_existing_images = parameters['existing_images'] == 'keep'

        # Make sure names in repository are valid images, and add tag if needed
        self.repositories = []
        for i, repository in enumerate(parameters['repository']):
            if is_image_name_id(repository):
                self.fail("repository[%d] must not be an image ID; got: %s" % (i + 1, repository))
            repo, repo_tag = parse_repository_tag(repository)
            if not repo_tag:
                repo_tag = parameters['tag']
            elif not is_valid_tag(repo_tag, allow_empty=False):
                self.fail("repository[%d] must not have a digest; got: %s" % (i + 1, repository))
            self.repositories.append((repo, repo_tag))

    def fail(self, msg):
        self.client.fail(msg)

    def tag_image(self, image, name, tag):
        tagged_image = self.client.find_image(name=name, tag=tag)
        if tagged_image:
            # Idempotency checks
            if tagged_image['Id'] == image['Id']:
                return (
                    False,
                    "target image already exists (%s) and is as expected" % tagged_image['Id'],
                    tagged_image,
                )
            if self.keep_existing_images:
                return (
                    False,
                    "target image already exists (%s) and is not as expected, but kept" % tagged_image['Id'],
                    tagged_image,
                )
            msg = "target image existed (%s) and was not as expected" % tagged_image['Id']
        else:
            msg = "target image did not exist"

        if not self.check_mode:
            try:
                params = {
                    'tag': tag,
                    'repo': name,
                    'force': True,
                }
                res = self.client._post(self.client._url('/images/{0}/tag', image['Id']), params=params)
                self.client._raise_for_status(res)
                if res.status_code != 201:
                    raise Exception("Tag operation failed.")
            except Exception as exc:
                self.fail("Error: failed to tag image as %s:%s - %s" % (name, tag, to_native(exc)))

        return True, msg, tagged_image

    def tag_images(self):
        if is_image_name_id(self.name):
            image = self.client.find_image_by_id(self.name, accept_missing_image=False)
        else:
            image = self.client.find_image(name=self.name, tag=self.tag)
            if not image:
                self.fail("Cannot find image %s:%s" % (self.name, self.tag))

        before = []
        after = []
        tagged_images = []
        results = dict(
            changed=False,
            actions=[],
            image=image,
            tagged_images=tagged_images,
            diff=dict(before=dict(images=before), after=dict(images=after)),
        )
        for repository, tag in self.repositories:
            tagged, msg, old_image = self.tag_image(image, repository, tag)
            before.append(image_info(repository, tag, old_image))
            after.append(image_info(repository, tag, image if tagged else old_image))
            if tagged:
                results['changed'] = True
                results['actions'].append('Tagged image %s as %s:%s: %s' % (image['Id'], repository, tag, msg))
                tagged_images.append('%s:%s' % (repository, tag))
            else:
                results['actions'].append('Not tagged image %s as %s:%s: %s' % (image['Id'], repository, tag, msg))

        return results


def main():
    argument_spec = dict(
        name=dict(type='str', required=True),
        tag=dict(type='str', default='latest'),
        repository=dict(type='list', elements='str', required=True),
        existing_images=dict(type='str', choices=['keep', 'overwrite'], default='overwrite'),
    )

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    try:
        results = ImageTagger(client).tag_images()
        client.module.exit_json(**results)
    except DockerException as e:
        client.fail('An unexpected Docker error occurred: {0}'.format(to_native(e)), exception=traceback.format_exc())
    except RequestException as e:
        client.fail(
            'An unexpected requests error occurred when trying to talk to the Docker daemon: {0}'.format(to_native(e)),
            exception=traceback.format_exc())


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Cindy Zhao (@cizhao) <cizhao@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_vrf_region_hub_network
short_description: Manage site-local VRF region hub network in schema template
description:
- Manage site-local VRF region hub network in schema template on Cisco ACI Multi-Site.
- The 'Hub Network' feature was introduced in Multi-Site Orchestrator (MSO) version 3.0(1) for AWS and version 3.0(2) for Azure.
author:
- Cindy Zhao (@cizhao)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  site:
    description:
    - The name of the site.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  vrf:
    description:
    - The name of the VRF.
    type: str
    required: true
  region:
    description:
    - The name of the region.
    type: str
    required: true
  hub_network:
    description:
    - The hub network to be managed.
    type: dict
    suboptions:
      name:
        description:
        - The name of the hub network.
        - The hub-default is the default created hub network.
        type: str
        required: true
      tenant:
        description:
        - The tenant name of the hub network.
        type: str
        required: true
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
notes:
- The ACI MultiSite PATCH API has a deficiency requiring some objects to be referenced by index.
  This can cause silent corruption on concurrent access when changing/removing on object as
  the wrong object may be referenced. This module is affected by this deficiency.
seealso:
- module: cisco.mso.mso_schema_site_vrf_region
- module: cisco.mso.mso_schema_template_vrf
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new site VRF region hub network
  cisco.mso.mso_schema_site_vrf_region_hub_network:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    hub_network:
      name: hub-default
      tenant: infra
    state: present
  delegate_to: localhost

- name: Remove a site VRF region hub network
  cisco.mso.mso_schema_site_vrf_region_hub_network:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    state: absent
  delegate_to: localhost

- name: Query site VRF region hub network
  cisco.mso.mso_schema_site_vrf_region_hub_network:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec, mso_hub_network_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        site=dict(type="str", required=True),
        template=dict(type="str", required=True),
        vrf=dict(type="str", required=True),
        region=dict(type="str", required=True),
        hub_network=dict(type="dict", options=mso_hub_network_spec()),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["hub_network"]],
        ],
    )

    schema = module.params.get("schema")
    site = module.params.get("site")
    template = module.params.get("template").replace(" ", "")
    vrf = module.params.get("vrf")
    region = module.params.get("region")
    hub_network = module.params.get("hub_network")
    state = module.params.get("state")

    mso = MSOModule(module)

    # Get schema objects
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get template
    templates = [t.get("name") for t in schema_obj.get("templates")]
    if template not in templates:
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template, ", ".join(templates)))

    # Get site
    site_id = mso.lookup_site(site)

    # Get site_idx
    if not schema_obj.get("sites"):
        mso.fail_json(msg="No site associated with template '{0}'. Associate the site with the template using mso_schema_site.".format(template))
    sites = [(s.get("siteId"), s.get("templateName")) for s in schema_obj.get("sites")]
    if (site_id, template) not in sites:
        mso.fail_json(msg="Provided site-template association '{0}-{1}' does not exist.".format(site, template))

    # Schema-access uses indexes
    site_idx = sites.index((site_id, template))
    # Path-based access uses site_id-template
    site_template = "{0}-{1}".format(site_id, template)

    # Get VRF
    vrf_ref = mso.vrf_ref(schema_id=schema_id, template=template, vrf=vrf)
    vrfs = [v.get("vrfRef") for v in schema_obj.get("sites")[site_idx]["vrfs"]]
    vrfs_name = [mso.dict_from_ref(v).get("vrfName") for v in vrfs]
    if vrf_ref not in vrfs:
        mso.fail_json(msg="Provided vrf '{0}' does not exist. Existing vrfs: {1}".format(vrf, ", ".join(vrfs_name)))
    vrf_idx = vrfs.index(vrf_ref)

    # Get Region
    regions = [r.get("name") for r in schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"]]
    if region not in regions:
        mso.fail_json(msg="Provided region '{0}' does not exist. Existing regions: {1}".format(region, ", ".join(regions)))
    region_idx = regions.index(region)
    # Get Region object
    region_obj = schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"][region_idx]
    region_path = "/sites/{0}/vrfs/{1}/regions/{2}".format(site_template, vrf, region)

    # Get hub network
    existing_hub_network = region_obj.get("cloudRsCtxProfileToGatewayRouterP")
    if existing_hub_network is not None:
        mso.existing = existing_hub_network

    if state == "query":
        if not mso.existing:
            mso.fail_json(msg="Hub network not found")
        mso.exit_json()

    ops = []

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=region_path + "/cloudRsCtxProfileToGatewayRouterP"))
            ops.append(dict(op="replace", path=region_path + "/isTGWAttachment", value=False))

    elif state == "present":
        new_hub_network = dict(
            name=hub_network.get("name"),
            tenantName=hub_network.get("tenant"),
        )
        payload = region_obj
        payload.update(
            cloudRsCtxProfileToGatewayRouterP=new_hub_network,
            isTGWAttachment=True,
        )

        mso.sanitize(payload, collate=True)

        ops.append(dict(op="replace", path=region_path, value=mso.sent))

        mso.existing = new_hub_network

    if not module.check_mode and mso.previous != mso.existing:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: discovery_job_info
short_description: Information module for Discovery Job
description:
- Get all Discovery Job.
- Get Discovery Job by id.
- >
   Returns the list of discovery jobs for the given Discovery ID. The results can be optionally filtered based on IP.
   Discovery ID can be obtained using the "Get Discoveries by range" API.
- Returns the list of discovery jobs for the given IP.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  offset:
    description:
    - Offset query parameter.
    type: int
  limit:
    description:
    - Limit query parameter.
    type: int
  ipAddress:
    description:
    - IpAddress query parameter.
    type: str
  name:
    description:
    - Name query parameter.
    type: str
  id:
    description:
    - Id path parameter. Discovery ID.
    type: str
requirements:
- dnacentersdk >= 2.5.5
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Discovery GetDiscoveryJobsByIP
  description: Complete reference of the GetDiscoveryJobsByIP API.
  link: https://developer.cisco.com/docs/dna-center/#!get-discovery-jobs-by-ip
- name: Cisco DNA Center documentation for Discovery GetListOfDiscoveriesByDiscoveryId
  description: Complete reference of the GetListOfDiscoveriesByDiscoveryId API.
  link: https://developer.cisco.com/docs/dna-center/#!get-list-of-discoveries-by-discovery-id
notes:
  - SDK Method used are
    discovery.Discovery.get_discovery_jobs_by_ip,
    discovery.Discovery.get_list_of_discoveries_by_discovery_id,

  - Paths used are
    get /dna/intent/api/v1/discovery/job,
    get /dna/intent/api/v1/discovery/{id}/job,

"""

EXAMPLES = r"""
- name: Get all Discovery Job
  cisco.dnac.discovery_job_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    offset: 0
    limit: 0
    ipAddress: string
    name: string
  register: result

- name: Get Discovery Job by id
  cisco.dnac.discovery_job_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    offset: 0
    limit: 0
    ipAddress: string
    id: string
  register: result

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": [
        {
          "attributeInfo": {},
          "cliStatus": "string",
          "discoveryStatus": "string",
          "endTime": "string",
          "httpStatus": "string",
          "id": "string",
          "inventoryCollectionStatus": "string",
          "inventoryReachabilityStatus": "string",
          "ipAddress": "string",
          "jobStatus": "string",
          "name": "string",
          "netconfStatus": "string",
          "pingStatus": "string",
          "snmpStatus": "string",
          "startTime": "string",
          "taskId": "string"
        }
      ],
      "version": "string"
    }
"""
